from terran import default_device
from terran.checkpoint import get_class_for_checkpoint

TASK_NAME = 'face-recognition'


class Recognition:

    def __init__(self, checkpoint=None, device=default_device, lazy=False):
        """Initializes and loads the model for `checkpoint`.

        Parameters
        ----------
        checkpoint : str
            Checkpoint (and model) to use in order to perform face recognition.
            If `None`, will use the default one for the task.
        device : torch.Device
            Device to load the model on.
        lazy : bool
            If set, will defer model loading until first call.

        """
        self.device = device
        self.recognition_cls = get_class_for_checkpoint(TASK_NAME, checkpoint)

        # Load the model into memory unless we have the lazy loading set.
        self.model = (
            self.recognition_cls(device=self.device) if not lazy else None
        )

    def __repr__(self):
        return f'<Recognition({self.recognition_cls.__name__})>'

    def __call__(self, images, faces_per_image=None):
        """Performs face recognition on `images`.

        Derives the actual prediction to the model the `Recognition` object was
        initialized with.

        Parameters
        ----------
        images : list of numpy.ndarray or numpy.ndarray
            Images to perform face recognition on.
        faces_per_image : list of list of dicts.
            Each dict entry must contain `bbox` and `landmarks` keys, as
            returned by a `Detection` instance.

        Returns
        -------
        list of numpy.ndarray or numpy.ndarray
            One entry per image, with a numpy array of size (N_i, F), with F
            being the embedding size returned by the model.

        """
        # If `images` is a single image, turn into a list.
        expanded = False
        if (
            not (isinstance(images, list) or isinstance(images, tuple))
            and len(images.shape) == 3
        ):
            expanded = True
            images = [images]

            # Also expand `faces_per_image`. If only received a `dict`, must
            # expand it twice.
            if isinstance(faces_per_image, dict):
                faces_per_image = [[faces_per_image]]
            else:
                faces_per_image = [faces_per_image]

        # Make sure that `images` and `faces_per_image`, if present, match
        # their ranks.
        if faces_per_image is not None and len(faces_per_image) != len(images):
            raise ValueError(
                f'`images` and `faces_per_image` must be of the same size, '
                f'but the former is of size {len(images)} while the latter of '
                f'size {len(faces_per_image)}.'
            )

        if self.model is None:
            self.model = self.recognition_cls(device=self.device)
        out = self.model.call(images, faces_per_image)

        # What we return depends on how much we expanded.
        if expanded and isinstance(faces_per_image, dict):
            return out[0][0]
        elif expanded:
            return out[0]
        else:
            return out


extract_features = Recognition(lazy=True)
"""Default entry point to face recognition.

This is an instantiation of the :class:`Recognition` class, lazily-loaded in
order to avoid reading the checkpoints on import. Refer to that class'
:func:`__call__ <Recognition.__call__>` method for more information.
"""



import torch

from collections import OrderedDict


def _make_layers(block, no_relu_layers):
    layers = []
    for layer_name, v in block.items():
        if 'pool' in layer_name:
            layer = torch.nn.MaxPool2d(
                kernel_size=v[0], stride=v[1], padding=v[2]
            )
            layers.append((layer_name, layer))
        else:
            conv2d = torch.nn.Conv2d(
                in_channels=v[0], out_channels=v[1],
                kernel_size=v[2], stride=v[3], padding=v[4]
            )
            layers.append((layer_name, conv2d))
            if layer_name not in no_relu_layers:
                layers.append(
                    ('relu_' + layer_name, torch.nn.ReLU(inplace=True))
                )
    return torch.nn.Sequential(OrderedDict(layers))


class BodyPoseModel(torch.nn.Module):

    def __init__(self):
        super(BodyPoseModel, self).__init__()

        no_relu_layers = [
            'conv5_5_CPM_L1', 'conv5_5_CPM_L2',
            'Mconv7_stage2_L1', 'Mconv7_stage2_L2',
            'Mconv7_stage3_L1', 'Mconv7_stage3_L2',
            'Mconv7_stage4_L1', 'Mconv7_stage4_L2',
            'Mconv7_stage5_L1', 'Mconv7_stage5_L2',
            'Mconv7_stage6_L1', 'Mconv7_stage6_L1'
        ]
        blocks = {}
        block0 = OrderedDict({
            'conv1_1': [3, 64, 3, 1, 1],
            'conv1_2': [64, 64, 3, 1, 1],
            'pool1_stage1': [2, 2, 0],
            'conv2_1': [64, 128, 3, 1, 1],
            'conv2_2': [128, 128, 3, 1, 1],
            'pool2_stage1': [2, 2, 0],
            'conv3_1': [128, 256, 3, 1, 1],
            'conv3_2': [256, 256, 3, 1, 1],
            'conv3_3': [256, 256, 3, 1, 1],
            'conv3_4': [256, 256, 3, 1, 1],
            'pool3_stage1': [2, 2, 0],
            'conv4_1': [256, 512, 3, 1, 1],
            'conv4_2': [512, 512, 3, 1, 1],
            'conv4_3_CPM': [512, 256, 3, 1, 1],
            'conv4_4_CPM': [256, 128, 3, 1, 1]
        })
        block1_1 = OrderedDict({
            'conv5_1_CPM_L1': [128, 128, 3, 1, 1],
            'conv5_2_CPM_L1': [128, 128, 3, 1, 1],
            'conv5_3_CPM_L1': [128, 128, 3, 1, 1],
            'conv5_4_CPM_L1': [128, 512, 1, 1, 0],
            'conv5_5_CPM_L1': [512, 38, 1, 1, 0]
        })
        block1_2 = OrderedDict({
            'conv5_1_CPM_L2': [128, 128, 3, 1, 1],
            'conv5_2_CPM_L2': [128, 128, 3, 1, 1],
            'conv5_3_CPM_L2': [128, 128, 3, 1, 1],
            'conv5_4_CPM_L2': [128, 512, 1, 1, 0],
            'conv5_5_CPM_L2': [512, 19, 1, 1, 0]
        })
        blocks['block1_1'] = block1_1
        blocks['block1_2'] = block1_2

        self.model0 = _make_layers(block0, no_relu_layers)

        for i in range(2, 7):
            blocks['block%d_1' % i] = OrderedDict({
                'Mconv1_stage%d_L1' % i: [185, 128, 7, 1, 3],
                'Mconv2_stage%d_L1' % i: [128, 128, 7, 1, 3],
                'Mconv3_stage%d_L1' % i: [128, 128, 7, 1, 3],
                'Mconv4_stage%d_L1' % i: [128, 128, 7, 1, 3],
                'Mconv5_stage%d_L1' % i: [128, 128, 7, 1, 3],
                'Mconv6_stage%d_L1' % i: [128, 128, 1, 1, 0],
                'Mconv7_stage%d_L1' % i: [128, 38, 1, 1, 0]
            })
            blocks['block%d_2' % i] = OrderedDict({
                'Mconv1_stage%d_L2' % i: [185, 128, 7, 1, 3],
                'Mconv2_stage%d_L2' % i: [128, 128, 7, 1, 3],
                'Mconv3_stage%d_L2' % i: [128, 128, 7, 1, 3],
                'Mconv4_stage%d_L2' % i: [128, 128, 7, 1, 3],
                'Mconv5_stage%d_L2' % i: [128, 128, 7, 1, 3],
                'Mconv6_stage%d_L2' % i: [128, 128, 1, 1, 0],
                'Mconv7_stage%d_L2' % i: [128, 19, 1, 1, 0]
            })

        for k in blocks.keys():
            blocks[k] = _make_layers(blocks[k], no_relu_layers)

        self.model1_1 = blocks['block1_1']
        self.model2_1 = blocks['block2_1']
        self.model3_1 = blocks['block3_1']
        self.model4_1 = blocks['block4_1']
        self.model5_1 = blocks['block5_1']
        self.model6_1 = blocks['block6_1']

        self.model1_2 = blocks['block1_2']
        self.model2_2 = blocks['block2_2']
        self.model3_2 = blocks['block3_2']
        self.model4_2 = blocks['block4_2']
        self.model5_2 = blocks['block5_2']
        self.model6_2 = blocks['block6_2']

    def forward(self, x):

        out1 = self.model0(x)

        out1_1 = self.model1_1(out1)
        out1_2 = self.model1_2(out1)
        out2 = torch.cat([out1_1, out1_2, out1], 1)

        out2_1 = self.model2_1(out2)
        out2_2 = self.model2_2(out2)
        out3 = torch.cat([out2_1, out2_2, out1], 1)

        out3_1 = self.model3_1(out3)
        out3_2 = self.model3_2(out3)
        out4 = torch.cat([out3_1, out3_2, out1], 1)

        out4_1 = self.model4_1(out4)
        out4_2 = self.model4_2(out4)
        out5 = torch.cat([out4_1, out4_2, out1], 1)

        out5_1 = self.model5_1(out5)
        out5_2 = self.model5_2(out5)
        out6 = torch.cat([out5_1, out5_2, out1], 1)

        out6_1 = self.model6_1(out6)
        out6_2 = self.model6_2(out6)

        return out6_1, out6_2



import numpy as np
import requests

from io import BytesIO
from pathlib import Path
from PIL import Image
from urllib.parse import urlparse


# Use a Chrome-based user agent to avoid getting needlessly blocked.
USER_AGENT = (
    'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) '
    'Chrome/51.0.2704.103 Safari/537.36'
)


def open_image(uri):
    """Opens and returns image at `uri`.

    Always returns a HxWxC `np.array` containing the image, ready to be
    consumed by any Terran algorithm. If the image is grayscale or has an alpha
    channel, it'll get converted into `RGB`, so the number of channels will
    always be 3.

    Parameters
    ----------
    uri : str or pathlib.Path
        URI pointing to an image, which may be a filesystem location or a URL.

    Returns
    -------
    numpy.ndarray
        Array of size HxWxC containing the pixel values for the image, as
        numpy.uint8.

    """
    # Check if `uri` is a URL or a filesystem location.
    if isinstance(uri, Path):
        image = Image.open(uri)
    elif urlparse(uri).scheme:
        response = requests.get(uri, headers={'User-Agent': USER_AGENT})
        image = Image.open(BytesIO(response.content))
    else:
        image = Image.open(Path(uri).expanduser())

    image = np.asarray(image.convert('RGB'))

    if len(image.shape) == 2:
        # Grayscale image, turn it to a rank-3 tensor anyways.
        image = np.stack([image] * 3, axis=-1)

    return image


def resolve_images(path, batch_size=None):
    """Collects the paths of all images under `path`, yielding them in batches.

    Ensures that the image is valid before returning it by attempting to open
    it with PIL.

    Parameters
    ----------
    path : str or pathlib.Path
        Path to recursively search for images in.

    Yields
    ------
    pathlib.Path or [pathlib.Path]
        Path to every valid image found under `path`. If `batch_size` is
        `None`, will return a single `pathlib.Path`. Otherwise, returns a list.

    """
    if not isinstance(path, Path):
        path = Path(path).expanduser()

    batch = []
    for f in path.glob('**/*'):
        if not f.is_file():
            continue

        try:
            Image.open(f).verify()
        except OSError:
            continue

        # If no `batch_size` specified, just return the path.
        if batch_size is None:
            yield path.joinpath(f)
            continue

        batch.append(path.joinpath(f))
        if len(batch) >= batch_size:
            yield batch
            batch = []



import math
import torch

from torchvision.ops import nms

from terran import default_device
from terran.checkpoint import get_checkpoint_path
from terran.face.detection.retinaface.anchors import (
    anchors_plane, generate_anchor_reference,
)
from terran.face.detection.retinaface.model import (
    RetinaFace as RetinaFaceModel
)


def load_model():
    model = RetinaFaceModel()
    model.load_state_dict(torch.load(
        get_checkpoint_path('terran.face.detection.retinaface.RetinaFace')
    ))
    model.eval()
    return model


def decode_bboxes(anchors, deltas):
    """Apply the bbox delta predictions on the base anchor coordinates.

    Paramters
    ---------
    anchors : torch.Tensor of shape (A, 4)
    deltas : torch.Tensor of shape (N, A, 4)

    Returns
    -------
    torch.Tensor of shape (N, A, 4)
        Adjusted bounding boxes.

    """
    widths = anchors[:, 2] - anchors[:, 0] + 1.0
    heights = anchors[:, 3] - anchors[:, 1] + 1.0
    ctr_x = anchors[:, 0] + 0.5 * (widths - 1.0)
    ctr_y = anchors[:, 1] + 0.5 * (heights - 1.0)

    dx = deltas[..., 0]  # (N, A)
    dy = deltas[..., 1]
    dw = deltas[..., 2]
    dh = deltas[..., 3]

    pred_ctr_x = dx * widths + ctr_x  # (N, A)
    pred_ctr_y = dy * heights + ctr_y
    pred_w = torch.exp(dw) * widths
    pred_h = torch.exp(dh) * heights

    # Perform the decoding in-place.
    pred = deltas
    pred[..., 0] = pred_ctr_x - 0.5 * (pred_w - 1.0)
    pred[..., 1] = pred_ctr_y - 0.5 * (pred_h - 1.0)
    pred[..., 2] = pred_ctr_x + 0.5 * (pred_w - 1.0)
    pred[..., 3] = pred_ctr_y + 0.5 * (pred_h - 1.0)

    return pred


def decode_landmarks(anchors, deltas):
    """Apply the landmark delta predictions on the base anchor coordinates.

    Paramters
    ---------
    anchors : torch.Tensor of shape (A, 4)
    deltas : torch.Tensor of shape (N, A, 5, 2)

    Returns
    -------
    torch.Tensor of shape (N, A, 5, 2)
        Adjusted landmark coordinates.

    """
    widths = anchors[:, 2] - anchors[:, 0] + 1.0
    heights = anchors[:, 3] - anchors[:, 1] + 1.0
    ctr_x = anchors[:, 0] + 0.5 * (widths - 1.0)
    ctr_y = anchors[:, 1] + 0.5 * (heights - 1.0)

    # Perform the decoding in-place.
    pred = deltas
    for i in range(5):
        pred[..., i, 0] = deltas[..., i, 0] * widths + ctr_x
        pred[..., i, 1] = deltas[..., i, 1] * heights + ctr_y

    return pred


class RetinaFace:

    def __init__(self, device=default_device, nms_threshold=0.4):
        self.device = device
        self.nms_threshold = nms_threshold

        # Anchor settings and feature strides used, specific to the `mnet`
        # backbone.
        self.feature_strides = [32, 16, 8]
        self.anchor_settings = {
            8: {
                'scales': (2, 1),
                'base_size': 16,
                'ratios': (1,),
            },
            16: {
                'scales': (8, 4),
                'base_size': 16,
                'ratios': (1,),
            },
            32: {
                'scales': (32, 16),
                'base_size': 16,
                'ratios': (1,),
            },
        }

        self.anchor_references = dict(zip(
            self.feature_strides,
            generate_anchor_reference(
                settings=self.anchor_settings, device=self.device
            )
        ))

        self.num_anchors_per_stride = {
            stride: anchors.shape[0]
            for stride, anchors in self.anchor_references.items()
        }

        self.model = load_model().to(self.device)

    def call(self, images, threshold=0.5):
        """Run the detection.

        `images` is a (N, H, W, C)-shaped array (np.float32).

        (Padding must be performed outside.)
        """
        H, W = images.shape[1:3]

        # Load the batch in to a `torch.Tensor` and pre-process by turning it
        # into a BGR format for the channels.
        data = torch.as_tensor(
            images, device=self.device, dtype=torch.float32
        ).permute(0, 3, 1, 2).flip(1)

        # Run the images through the network. Disable gradients, as they're not
        # needed.
        with torch.no_grad():
            output = self.model(data)

        # Calculate the base anchor coordinates per stride of the FPN.
        anchors_per_stride = {}
        for stride in self.feature_strides:
            # Input dimensions after model downsampling (i.e. feature map
            # dimensions).
            height = math.ceil(H / stride)
            width = math.ceil(W / stride)

            anchor_ref = self.anchor_references[stride]
            anchors = anchors_plane(anchor_ref, height, width, stride)
            anchors_per_stride[stride] = anchors

        # Decode the outputs of the model, adjusting the anchors.
        proposals_list = []
        scores_list = []
        landmarks_list = []
        for stride_idx, stride in enumerate(self.feature_strides):
            # Three per stride: class, bbox, landmark.
            idx = stride_idx * 3

            anchors = anchors_per_stride[stride]
            A = self.num_anchors_per_stride[stride]
            N = output[idx].shape[0]

            scores = output[idx]
            scores = scores[:, A:, :, :]
            scores = scores.permute(0, 2, 3, 1).reshape(N, -1)

            bbox_deltas = output[idx + 1]
            bbox_pred_len = bbox_deltas.shape[1] // A
            bbox_deltas = bbox_deltas.permute(0, 2, 3, 1).reshape(
                (N, -1, bbox_pred_len)
            )

            landmark_deltas = output[idx + 2]
            landmark_pred_len = landmark_deltas.shape[1] // A
            landmark_deltas = landmark_deltas.permute(0, 2, 3, 1).reshape(
                (N, -1, 5, landmark_pred_len // 5)
            )

            proposals = decode_bboxes(anchors, bbox_deltas)
            landmarks = decode_landmarks(anchors, landmark_deltas)

            scores_list.append(scores)
            proposals_list.append(proposals)
            landmarks_list.append(landmarks)

        batch_scores = torch.cat(scores_list, dim=1)
        batch_proposals = torch.cat(proposals_list, dim=1)
        batch_landmarks = torch.cat(landmarks_list, dim=1)

        # Collect the predictions per image, filtering low-scoring proposals
        # and performing NMS.
        batch_objects = []
        for image_idx in range(images.shape[0]):
            scores = batch_scores[image_idx]
            proposals = batch_proposals[image_idx]
            landmarks = batch_landmarks[image_idx]

            order = torch.where(scores >= threshold)[0]
            proposals = proposals[order, :]
            scores = scores[order]
            landmarks = landmarks[order, :]

            if proposals.shape[0] == 0:
                batch_objects.append([])
                continue

            # Re-order all proposals according to score.
            order = scores.argsort(descending=True)
            proposals = proposals[order]
            scores = scores[order]
            landmarks = landmarks[order]

            # Run the predictions through non-maximum suppression.
            keep = nms(proposals, scores, self.nms_threshold)
            proposals = proposals[keep].to('cpu').numpy()
            scores = scores[keep].to('cpu').numpy()
            landmarks = landmarks[keep].to('cpu').numpy()

            batch_objects.append([
                {'bbox': b, 'landmarks': l, 'score': s}
                for s, b, l in zip(scores, proposals, landmarks)
            ])

        return batch_objects



import math
import numpy as np

from cv2 import resize, INTER_LINEAR

from terran import default_device
from terran.checkpoint import get_class_for_checkpoint


TASK_NAME = 'face-detection'


def resize_factory(short_side=416):

    def resize_in(images):
        # If input is a numpy array, we simply modify all images at once with a
        # single scale, returning the result also as an ndarray. If the input,
        # however, is a list, we must resize each image individually, keeping
        # the scales for each.
        if isinstance(images, np.ndarray):
            H, W = images.shape[1:3]
            scale = short_side / min(H, W)

            new_size = (
                int(W * scale), int(H * scale)
            )

            resized = np.empty(
                (images.shape[0], new_size[1], new_size[0], images.shape[3]),
                dtype=images.dtype
            )
            for idx, image in enumerate(images):
                resize(
                    src=image,
                    dst=resized[idx],
                    dsize=new_size,
                    interpolation=INTER_LINEAR
                )
            scales = scale
        else:
            resized = []
            scales = []
            for image in images:
                H, W = image.shape[0:2]
                scale = short_side / min(H, W)
                new_size = (int(W * scale), int(H * scale))

                resized.append(
                    resize(
                        src=image,
                        dsize=new_size,
                        interpolation=INTER_LINEAR
                    )
                )
                scales.append(scale)

        return resized, scales

    def resize_out(faces_per_image, scales):
        # If we receive a single scale for all images, turn it into a scale per
        # image.
        if not isinstance(scales, list):
            scales = [scales] * len(faces_per_image)

        new_faces_per_image = []
        for faces, scale in zip(faces_per_image, scales):

            new_faces = []
            # TODO: Move rounding to a separate preprocessor. (To also ensure
            # coordinates are valid.)
            for face in faces:
                new_faces.append({
                    'bbox': np.around(
                        face['bbox'] / scale
                    ).astype(np.int32),
                    'landmarks': np.around(
                        face['landmarks'] / scale
                    ).astype(np.int32),
                    'score': face['score'],
                })

            new_faces_per_image.append(new_faces)

        return new_faces_per_image

    return resize_in, resize_out


def merge_factory(method='padding'):
    """Merges a list of images into a single array.

    May do so by padding or cropping images. See `Detection` for more
    information on the options.
    """

    def merge_in(images):
        # Check if not merged already and return early.
        if isinstance(images, np.ndarray):
            return images, {'merged': False}

        # if len(images) == 1:
        #     return np.array(images), {'merged': False}

        params = {'merged': True}

        if method == 'crop':
            raise NotImplementedError
        elif method == 'padding':
            max_height = max([arr.shape[0] for arr in images])
            max_width = max([arr.shape[1] for arr in images])
            padded = np.zeros(
                (len(images), max_height, max_width, 3),
                dtype=np.uint8
            )

            pads_per_image = []
            for idx, image in enumerate(images):
                diff_height = max(0, (max_height - image.shape[0]) / 2)
                diff_width = max(0, (max_width - image.shape[1]) / 2)
                pad_values = [
                    (
                        int(math.ceil(diff_height)),
                        int(math.floor(diff_height))
                    ),
                    (
                        int(math.ceil(diff_width)),
                        int(math.floor(diff_width))
                    ),
                    (0, 0),
                ]
                padded[idx, ...] = np.pad(image, pad_values)
                pads_per_image.append(pad_values)

            params['pads_per_image'] = pads_per_image
            return padded, params
        else:
            raise ValueError(
                'Invalid `method` set, options are `padding` or `crop`.'
            )

    def merge_out(faces_per_image, params):
        # If no merging occur, we have nothing to adjust.
        if not params['merged']:
            return faces_per_image

        if method == 'crop':
            raise NotImplementedError
        elif method == 'padding':
            new_faces_per_image = []
            for faces, pads in zip(faces_per_image, params['pads_per_image']):

                new_faces = []
                for face in faces:
                    bbox = np.array([
                        face['bbox'][0] - pads[1][0],
                        face['bbox'][1] - pads[0][0],
                        face['bbox'][2] - pads[1][0],
                        face['bbox'][3] - pads[0][0],
                    ])

                    pads_per_axis = np.array([
                        pads[1][0], pads[0][0]
                    ]).reshape(1, -1)
                    landmarks = (
                        face['landmarks'] - pads_per_axis
                    )

                    new_faces.append({
                        'bbox': bbox,
                        'landmarks': landmarks,
                        'score': face['score'],
                    })

                new_faces_per_image.append(new_faces)

            return new_faces_per_image
        else:
            raise ValueError(
                'Invalid `method` set, options are `padding` or `crop`.'
            )

    return merge_in, merge_out


class Detection:

    def __init__(
        self, checkpoint=None, short_side=416, merge_method='padding',
        device=default_device, lazy=False
    ):
        """Initializes and loads the model for `checkpoint`.

        Parameters
        ----------
        checkpoint : str
            Checkpoint (and model) to use in order to perform face detection.
            If `None`, will use the default one for the task.
        short_side : int
            Resize images' short side to `short_side` before sending over to
            detection model.
        merge_method : 'padding', 'crop'
            How to merge images together into a batch when receiving a
            list. Merge is done after resizing. Options are:

            * `padding`, which will add padding around images, possibly
              increasing total image size. If mixing portrait and landscape
              images, might be inefficient.

            * `crop`, which will center-crop the images to the smallest
              size. If images are of very different sizes, might end up
              cropping too much.

        device : torch.Device
            Device to load the model on.
        lazy : bool
            If set, will defer model loading until first call.

        """
        self.device = device
        self.detection_cls = get_class_for_checkpoint(TASK_NAME, checkpoint)

        # Load the model into memory unless we have the lazy loading set.
        self.model = (
            self.detection_cls(device=self.device) if not lazy else None
        )

        # TODO: Allow `max_size`, or by area.
        self.resize_in, self.resize_out = resize_factory(short_side=short_side)
        self.merge_in, self.merge_out = merge_factory(method=merge_method)

    def __repr__(self):
        return f'<Detection({self.detection_cls.__name__})>'

    def __call__(self, images):
        """Performs face detection on `images`.

        Derives the actual prediction to the model the `Detection` object was
        initialized with.

        Parameters
        ----------
        images : list of numpy.ndarray or numpy.ndarray
            Images to perform face detection on.

        Returns
        -------
        list of list of dicts, or list dict
            List of dictionaries containing face data for a single image, or a
            list of these entries thereof.

            Each entry is of the form::

                {
                    'bbox': [x_min, y_min, x_max, y_max],
                    'landmarks': ...,  # Array of shape (5, 2).
                    'score': ... # Confidence score.
                }

        """
        # If `images` is a single `np.ndarray`, turn into a list.
        expanded = False
        if (
            not (isinstance(images, list) or isinstance(images, tuple))
            and len(images.shape) == 3
        ):
            expanded = True
            images = np.expand_dims(images, 0)

        # Run generic image preprocessing, such as resizing images to
        # reasonable sizes.
        images, resize_params = self.resize_in(images)
        images, merge_params = self.merge_in(images)

        # Call the actual model on the images. If we haven't loaded the model
        # yet due to lazy loading, load it.
        if self.model is None:
            self.model = self.detection_cls(device=self.device)
        out = self.model.call(images)

        # Run generic result postprocessing, such as adjusting the coordinates
        # for rescaled images.
        # TODO: Add final validation transformer to check that all bounding
        # boxes are valid (larger than zero dimensions and clipped to image).
        out = self.merge_out(out, merge_params)
        out = self.resize_out(out, resize_params)

        return out[0] if expanded else out


face_detection = Detection(lazy=True)
"""Default entry point to face detection.

This is an instantiation of the :class:`Detection` class, lazily-loaded in
order to avoid reading the checkpoints on import. Refer to that class'
:func:`__call__ <Detection.__call__>` method for more information.
"""



import torch.nn as nn


class Unit(nn.Module):

    def __init__(self, in_c, out_c, stride=1):
        super().__init__()

        self.dimensions_match = in_c == out_c and stride == 1

        self.body = nn.Sequential(
            nn.BatchNorm2d(in_c, momentum=0.9, eps=2e-5),

            nn.Conv2d(in_c, out_c, 3, padding=1, bias=False),
            nn.BatchNorm2d(out_c, momentum=0.9, eps=2e-5),
            nn.PReLU(num_parameters=out_c),

            nn.Conv2d(out_c, out_c, 3, stride=stride, padding=1, bias=False),
            nn.BatchNorm2d(out_c, momentum=0.9, eps=2e-5),
        )

        # If input and output dimensions within unit don't match, apply a
        # convolutional layer to adjust the number of channels.
        if not self.dimensions_match:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_c, out_c, 1, stride=stride, bias=False),
                nn.BatchNorm2d(out_c, momentum=0.9, eps=2e-5),
            )
        else:
            self.shortcut = nn.Identity()

    def forward(self, x):
        body = self.body(x)
        shortcut = self.shortcut(x)
        return body + shortcut


class FaceResNet100(nn.Module):

    def __init__(self):
        super().__init__()

        # Network hyperparameters.
        self.units_per_stage = [3, 13, 30, 3]
        self.channels = [64, 64, 128, 256, 512]

        # Preprocessing parameters.
        self.mean = 127.5
        self.std = 0.0078125

        # Initial layers applied.
        self.initial_layer = nn.Sequential(
            nn.Conv2d(3, self.channels[0], 3, padding=1, bias=False),
            nn.BatchNorm2d(self.channels[0], momentum=0.9, eps=2e-5),
            nn.PReLU(num_parameters=self.channels[0]),
        )

        stages = []
        for stage_idx, num_units in enumerate(self.units_per_stage):
            # Number of channels for previous and current stage.
            prev_c = self.channels[stage_idx]
            curr_c = self.channels[stage_idx + 1]
            num_units = self.units_per_stage[stage_idx]

            units = nn.Sequential(*[
                Unit(prev_c, curr_c, stride=2),
                *[
                    Unit(curr_c, curr_c)
                    for _ in range(num_units - 1)
                ]
            ])

            stages.append(units)

        self.stages = nn.ModuleList(stages)

        # Final FC size is 7x7x512: 7x7 produced by an initial size of 112x112
        # and a downsampling rate of 16 of the Resnet, and 512 channels.
        self.final_layer = nn.Sequential(
            nn.BatchNorm2d(self.channels[-1], momentum=0.9, eps=2e-5),
            nn.Dropout(0.4),
            nn.Flatten(),
            nn.Linear(7 * 7 * 512, 512),
            nn.BatchNorm1d(512, momentum=0.9, eps=2e-5),
        )

    def forward(self, x):
        preprocessed = (x - self.mean) * self.std

        out = self.initial_layer(preprocessed)

        for stage in self.stages:
            out = stage(out)

        out = self.final_layer(out)

        return out



import torch
import numpy as np

from terran import default_device


def anchors_plane(anchor_ref, feat_h, feat_w, stride):
    """Builds the anchors plane for the given feature map shape.

    Based on an anchor reference, reproduces it at every point of the feature
    map, according to the specified stride. `anchor_ref` is set in real-image
    coordinates, so we must know what the stride for the current feature map
    is.

    Parameters
    ----------
    anchor_ref : torch.Tensor of size (A, 4)
        Coordinates for each of the `A` anchors, centered at the origin.
    feat_h : int
        Height of the feature map.
    feat_w : int
        Width of the feature map.
    stride : int
        Number of pixels every which to apply the anchor reference.

    Returns
    -------
    torch.Tensor of size (feat_h * feat_w * A, 4)
        The dtype and device of the returned tensor is based on `anchor_ref`.

    """
    device = anchor_ref.device
    dtype = anchor_ref.dtype

    shift_y, shift_x = torch.meshgrid(
        torch.arange(feat_h, dtype=dtype, device=device) * stride,
        torch.arange(feat_w, dtype=dtype, device=device) * stride,
    )

    shift_x = shift_x.reshape(-1)
    shift_y = shift_y.reshape(-1)

    shifts = torch.stack([
        shift_x, shift_y, shift_x, shift_y,
    ], dim=-1)

    anchors = (
        anchor_ref[None, ...] + shifts[:, None, :]
    ).reshape(-1, 4)

    return anchors


def generate_anchor_reference(settings=None, device=default_device):
    """Generate anchor reference per stride, for the given settings."""
    feature_strides = sorted(settings.keys(), reverse=True)

    anchors = []
    for stride in feature_strides:
        base_size = settings[stride]["base_size"]
        ratios = np.array(settings[stride]["ratios"])
        scales = np.array(settings[stride]["scales"])

        anchor = torch.as_tensor(
            generate_anchors(base_size, ratios, scales, stride),
            dtype=torch.float32,
            device=default_device
        )

        anchors.append(anchor)

    return anchors


def generate_anchors(base_size, ratios, scales, stride):
    """Generate an anchor reference for the given properties."""
    base_anchor = np.array([0, 0, base_size - 1, base_size - 1])
    ratio_anchors = _ratio_enum(base_anchor, ratios)
    anchors = np.vstack([
        _scale_enum(ratio_anchors[i, :], scales)
        for i in range(ratio_anchors.shape[0])
    ])

    return anchors


def _whctrs(anchor):
    """Return width, height, x center, and y center for an anchor (window)."""
    w = anchor[2] - anchor[0] + 1
    h = anchor[3] - anchor[1] + 1
    ctr_x = anchor[0] + 0.5 * (w - 1)
    ctr_y = anchor[1] + 0.5 * (h - 1)
    return w, h, ctr_x, ctr_y


def _mkanchors(ws, hs, ctr_x, ctr_y):
    """Given a vector of widths (ws) and heights (hs) around a center
    (ctr_x, ctr_y), output a set of anchors (windows).
    """
    ws = ws[:, np.newaxis]
    hs = hs[:, np.newaxis]

    anchors = np.hstack([
        ctr_x - 0.5 * (ws - 1),
        ctr_y - 0.5 * (hs - 1),
        ctr_x + 0.5 * (ws - 1),
        ctr_y + 0.5 * (hs - 1),
    ])
    return anchors


def _ratio_enum(anchor, ratios):
    """Enumerate a set of anchors for each aspect ratio wrt an anchor."""
    w, h, ctr_x, ctr_y = _whctrs(anchor)
    size = w * h
    size_ratios = size / ratios

    ws = np.round(np.sqrt(size_ratios))
    hs = np.round(ws * ratios)

    anchors = _mkanchors(ws, hs, ctr_x, ctr_y)

    return anchors


def _scale_enum(anchor, scales):
    """Enumerate a set of anchors for each scale wrt an anchor."""
    w, h, ctr_x, ctr_y = _whctrs(anchor)
    ws = w * scales
    hs = h * scales

    anchors = _mkanchors(ws, hs, ctr_x, ctr_y)

    return anchors



import click
import importlib
import os
import requests
import shutil
import sys
import tempfile

from itertools import groupby
from pathlib import Path


__all__ = [
    'get_terran_home',
    'get_class_for_checkpoint',
    'get_checkpoint_path',
]


DEFAULT_TERRAN_HOME = Path('~/.terran')
CHECKPOINT_PATH = 'checkpoints'

LABELS_BY_TASK = {
    'face-detection': 'Face detection (`terran.face.Detection`)',
    'face-recognition': 'Face recognition (`terran.face.Recognition`)',
    'pose-estimation': 'Pose estimation (`terran.pose.Estimation`)',
}

CHECKPOINTS = [
    {
        'id': 'b5d77fff',
        'name': 'RetinaFace',
        'description': (
            'RetinaFace with mnet backbone.'
        ),

        'task': 'face-detection',
        'class': 'terran.face.detection.retinaface.RetinaFace',
        'alias': 'gpu-realtime',
        'default': True,

        'performance': 1.0,
        'evaluation': {
            'value': 0.76,
            'metric': 'mAP',
            'is_reported': False,
        },

        'url': (
            'https://github.com/nagitsu/terran/releases/download/0.0.1/'
            'retinaface-mnet.pth'
        )
    },
    {
        'id': 'd206e4b0',
        'name': 'ArcFace',
        'description': (
            'ArcFace with Resnet 100 backbone.'
        ),

        'task': 'face-recognition',
        'class': 'terran.face.recognition.arcface.ArcFace',
        'alias': 'gpu-realtime',
        'default': True,

        'performance': 0.9,
        'evaluation': {
            'value': 0.80,
            'metric': 'accuracy',
            'is_reported': False,
        },

        'url': (
            'https://github.com/nagitsu/terran/releases/download/0.0.1/'
            'arcface-resnet100.pth'
        )
    },
    {
        'id': '11a769ad',
        'name': 'OpenPose',
        'description': (
            'OpenPose with VGG backend, 2017 version. Has some modifications, '
            'improving computational efficiency by giving up mAP.'
        ),

        'task': 'pose-estimation',
        'class': 'terran.pose.openpose.OpenPose',
        'alias': 'gpu-realtime',
        'default': True,

        'performance': 1.8,
        'evaluation': {
            'value': 0.65,
            'metric': 'mAP',
            'is_reported': True,
        },

        'url': (
            'https://github.com/nagitsu/terran/releases/download/0.0.1/'
            'openpose-body.pth'
        )
    },
]


def get_terran_home(create_if_missing=True):
    """Returns Terran's homedir.

    Defaults to `DEFAULT_TERRAN_HOME`, which is `~/.terran`, but can be
    overridden with the `TERRAN_HOME` environment variable.

    Returns
    -------
    pathlib.Path
        Path pointing to the base Terran directory.

    """
    path = Path(
        os.environ.get('TERRAN_HOME', DEFAULT_TERRAN_HOME)
    ).expanduser()

    # Create the directory if it doesn't exist.
    if create_if_missing and not path.exists():
        path.mkdir(exist_ok=True)

    return path


def get_checkpoints_directory():
    """Returns checkpoint directory within Terran's homedir.

    If the path doesn't exists, creates it.

    Returns
    -------
    pathlib.Path
        Path pointing to the checkpoints directory.

    """
    path = get_terran_home() / CHECKPOINT_PATH
    path.mkdir(exist_ok=True)
    return path


def read_checkpoint_db():
    """Reads the checkpoints database file from disk."""
    # Get the downloaded checkpoints by searching through the filesystem.
    local_checkpoints = set(
        path.stem for path in get_checkpoints_directory().glob('*.pth')
    )

    checkpoints = [
        {
            'status': (
                'DOWNLOADED' if checkpoint['id'] in local_checkpoints
                else 'NOT_DOWNLOADED'
            ),
            'local_path': (
                get_checkpoints_directory() / f"{checkpoint['id']}.pth"
                if checkpoint['id'] in local_checkpoints else None
            ),
            **checkpoint
        }
        for checkpoint in CHECKPOINTS
    ]

    return {
        'checkpoints': checkpoints
    }


def get_checkpoint(db, id_or_alias):
    """Returns checkpoint entry in `db` indicated by `id_or_alias`.

    Parameters
    ----------
    id_or_alias : str or tuple of str
        Either the ID of the checkpoint, or a tuple of the form (task_name,
        alias) that uniquely points to a checkpoint.

        In the case of the latter, specifying `alias` as `None` will default to
        the entry marked as `default = True` in the index.

    Returns
    -------
    Dict
        Checkpoint data contained in the database.

    """
    if isinstance(id_or_alias, tuple):
        task_name, alias = id_or_alias
        selected = [
            c for c in db['checkpoints']
            if c['task'] == task_name and (
                c['alias'] == alias if alias is not None else c['default']
            )
        ]
    else:
        selected = [c for c in db['checkpoints'] if c['id'] == id_or_alias]

    if len(selected) < 1:
        return None

    if len(selected) > 1:
        click.echo(
            f"Multiple checkpoints found for '{id_or_alias}' "
            f"({len(selected)}). Returning first."
        )

    return selected[0]


def get_class_for_checkpoint(task_name, alias):
    """Returns the model class for the given ID or alias.

    Parameters
    ----------
    task_name : str
        Name of the task to get the class for.
    alias : str or `None`
        Alias for the checkpoint, within `task_name`, to get the class for.
        If `None`, will use the first entry that specifies `default = True`.

    Returns
    -------
    class
        The class of the model associated to the checkpoint.

    Raises
    ------
    ValueError
        If checkpoint is not found.
    ImportError
        If the class is somehow misspecified or not present.

    """
    db = read_checkpoint_db()
    checkpoint = get_checkpoint(db, (task_name, alias))

    if not checkpoint:
        # No checkpoint found.
        raise ValueError('Checkpoint not found.')

    module_path, class_name = checkpoint['class'].rsplit('.', maxsplit=1)
    return getattr(importlib.import_module(module_path), class_name)


def get_checkpoint_by_class(db, class_path):
    """Returns checkpoint entry in `db` indicated by `class_path`.

    Parameters
    ----------
    class_path : str
        Fully specified path to class (e.g. `terran.pose.openpose.OpenPose`)
        of the model to get the checkpoint for.

    Returns
    -------
    Dict
        Checkpoint data contained in the database.

    """
    selected = [c for c in db['checkpoints'] if c['class'] == class_path]

    if len(selected) < 1:
        return None

    if len(selected) > 1:
        click.echo(
            f"Multiple checkpoints found for '{class_path}' "
            f"({len(selected)}). Returning first."
        )

    return selected[0]


def get_checkpoint_path(model_class_path, prompt=True):
    """Returns the local path to the model's weights.

    Goes through the list of checkpoints and returns the local path to the
    weights of the modell specified by `model_class`. If the weights are not
    downloaded, downloads them first.

    Parameters
    ----------
    model_class_path : str
        Fully specified path to class (e.g. `terran.pose.openpose.OpenPose`)
        of the model to get the checkpoint for.
    prompt : boolean
        If `True` and the checkpoint is not yet downloaded, prompt to download.

    Returns
    -------
    pathlib.Path
        Path to the `.pth` file containing the weights for the model.

    Raises
    ------
    ValueError
        If checkpoint is not found or is found but not downloaded, either due
        to aborting the prompt or disabling it in the first place.

    """
    db = read_checkpoint_db()
    checkpoint = get_checkpoint_by_class(db, model_class_path)
    can_prompt = sys.stdout.isatty()

    if not checkpoint:
        # No checkpoint found.
        raise ValueError('Checkpoint not found.')

    if checkpoint['status'] == 'NOT_DOWNLOADED':
        if prompt and can_prompt:
            # Checkpoint hasn't been downloaded yet. Prompt for downloading it
            # before continuing.
            try:
                click.confirm(
                    'Checkpoint not present locally. Want to download it?', abort=True
                )
            except Exception:
                click.echo('Checkpoint not present locally. Downloading it')

        download_remote_checkpoint(db, checkpoint)

    # `local_path` is now present in the checkpoint, as
    # `download_remote_checkpoint` will have modified it if it was just
    # downloaded.
    return checkpoint['local_path']


def download_remote_checkpoint(db, checkpoint):
    # Check if the checkpoint is already downloaded. If it is, this function
    # shouldn't have been called, so something is inconsistent.
    if checkpoint['local_path'] and checkpoint['local_path'].exists():
        click.echo(
            f"Checkpoint file already present at {checkpoint['local_path']}. '"
            f"If you're running into any issues, try issuing a `terran "
            f"checkpoint delete {checkpoint['id']}` trying attempting again."
        )
        return

    file_name = f"{checkpoint['id']}.pth"

    # Create a temporary directory to download the files into.
    tempdir = tempfile.mkdtemp()
    path = Path(tempdir) / file_name

    # Start the actual file download.
    response = requests.get(checkpoint['url'], stream=True)

    if response.status_code != 200:
        raise ValueError(f'Invalid checkpoint URL {checkpoint["url"]}')

    length = int(response.headers.get('Content-Length'))
    chunk_size = 16 * 1024
    progressbar = click.progressbar(
        response.iter_content(chunk_size=chunk_size),
        length=length / chunk_size, label='Downloading checkpoint...',
    )

    with open(path, 'wb') as f:
        with progressbar as content:
            for chunk in content:
                f.write(chunk)

    # Move the file to the checkpoints directory.
    new_path = get_checkpoints_directory() / file_name
    shutil.move(path, new_path)

    # Update the checkpoint dict information.
    checkpoint['status'] = 'DOWNLOADED'
    checkpoint['local_path'] = new_path

    # And finally make sure to delete the temp dir.
    shutil.rmtree(tempdir)

    click.echo("Checkpoint downloaded successfully.")


@click.command(name='list', help='List available checkpoints.')
def list_cmd():
    db = read_checkpoint_db()

    if not db['checkpoints']:
        click.echo('No checkpoints available.')
        return

    template = '| {:>30} | {:>12} | {:>8} | {:>8} | {:>14} |'

    header = template.format(
        'Name', 'Alias', 'Eval.', 'Perf.', 'Status'
    )
    click.echo('=' * len(header))
    click.echo(header)

    is_first = True
    for key, group in groupby(db['checkpoints'], key=lambda x: x['task']):
        label = LABELS_BY_TASK.get(key, '')
        click.echo(('=' if is_first else '-') * len(header))
        click.echo(f'| {label:<{len(header) - 4}} |')
        click.echo('-' * len(header))
        is_first = False

        for checkpoint in group:
            line = template.format(
                f"{checkpoint['name']} ({checkpoint['id']})",
                checkpoint['alias'],
                '',
                '',
                # (
                #     ('* ' if checkpoint['evaluation']['is_reported'] else '')
                #     + f"{checkpoint['evaluation']['value']:.2f}"
                # ),
                # f"{checkpoint['performance']:.2f}",
                checkpoint['status'],
            )
            click.echo(line)

    click.echo('=' * len(header))


@click.command(name='info', help='Display detailed information on checkpoint.')
@click.argument('checkpoint_id')
def info_cmd(checkpoint_id):
    db = read_checkpoint_db()

    checkpoint = get_checkpoint(db, checkpoint_id)
    if not checkpoint:
        click.echo(
            "Checkpoint '{}' not found in index.".format(checkpoint_id)
        )
        return

    click.echo(
        f"{checkpoint['name']} ({checkpoint['id']}, {checkpoint['alias']})"
    )

    if checkpoint['description']:
        click.echo(f" > {checkpoint['description']}")

    click.echo()

    click.echo(f"Task: {LABELS_BY_TASK.get(checkpoint['task'], '')}")
    click.echo(f"Class: `{checkpoint['class']}`")
    click.echo('Evaluation information: {:.3f} {}{}'.format(
        checkpoint['evaluation']['value'],
        checkpoint['evaluation']['metric'],
        ' (self-reported)' if checkpoint['evaluation']['is_reported'] else ''
    ))
    click.echo(
        f"Computational performance: {checkpoint['performance']:.2f} units"
    )

    click.echo()

    click.echo(f"Upstream URL: {checkpoint['url']}")
    if checkpoint['local_path']:
        click.echo(f"Status: DOWNLOADED (at `{checkpoint['local_path']}`)")
    else:
        click.echo('Status: NOT_DOWNLOADED')


@click.command(
    name='delete', help='Delete local files associated to a checkpoint.'
)
@click.argument('checkpoint_id')
def delete_cmd(checkpoint_id):
    db = read_checkpoint_db()
    checkpoint = get_checkpoint(db, checkpoint_id)
    if not checkpoint:
        click.echo(f"Checkpoint `{checkpoint_id}` not found in index.")
        return

    if checkpoint['status'] == 'NOT_DOWNLOADED':
        click.echo("Checkpoint isn't downloaded. Nothing to delete.")
        return

    # Delete the files associated to checkpoint.
    checkpoint['local_path'].unlink()

    click.echo(f"Checkpoint `{checkpoint['id']}` deleted successfully.")


@click.command(name='download', help='Download a remote checkpoint.')
@click.argument('checkpoint_id')
def download_cmd(checkpoint_id):
    db = read_checkpoint_db()
    checkpoint = get_checkpoint(db, checkpoint_id)
    if not checkpoint:
        click.echo(f"Checkpoint `{checkpoint_id}` not found in index.")
        return

    if checkpoint['status'] != 'NOT_DOWNLOADED':
        click.echo("Checkpoint is already downloaded.")
        return

    download_remote_checkpoint(db, checkpoint)


@click.group(name='checkpoint', help='Checkpoint management commands.')
def checkpoint_cmd():
    pass


checkpoint_cmd.add_command(delete_cmd)
checkpoint_cmd.add_command(download_cmd)
checkpoint_cmd.add_command(info_cmd)
checkpoint_cmd.add_command(list_cmd)



"""Module containig I/O-related functions for video handling."""
# Default size for the reader buffer, in batches. A size of 1, which is
# equivalent to prefetching the following batch, seems to be more than enough,
# and doesn't use up too much extra memory (~190MB for a 1080p batch of size
# 32).
DEFAULT_READER_BUFFER_SIZE = 1

# Default size for the writer buffer, in frames.
DEFAULT_WRITER_BUFFER_SIZE = 64


class EndOfVideo(Exception):
    pass


class VideoClosed(Exception):
    pass


from terran.io.video.reader import open_video  # noqa
from terran.io.video.writer import write_video  # noqa



import math
import numpy as np

from enum import Enum

from terran import default_device
from terran.checkpoint import get_class_for_checkpoint


TASK_NAME = 'pose-estimation'


class Keypoint(Enum):
    NOSE = 0
    NECK = 1

    R_SHOULDER = 2
    R_ELBOW = 3
    R_HAND = 4

    L_SHOULDER = 5
    L_ELBOW = 6
    L_HAND = 7

    R_HIP = 8
    R_KNEE = 9
    R_FOOT = 10

    L_HIP = 11
    L_KNEE = 12
    L_FOOT = 13

    R_EYE = 14
    L_EYE = 15
    R_EAR = 16
    L_EAR = 17


# TODO: This is *almost* like `terran.face.detection.merge_factory`. Unify them
# so we don't repeat so much code.
def merge_factory(method='padding'):
    """Merges a list of images into a single array.

    May do so by padding or cropping images. See `Estimation` for more
    information on the options.
    """

    def merge_in(images):
        # Check if not merged already and return early.
        if isinstance(images, np.ndarray):
            return images, {'merged': False}

        # TODO: Why is this commented?
        # if len(images) == 1:
        #     return np.array(images), {'merged': False}

        params = {'merged': True}

        if method == 'crop':
            raise NotImplementedError
        elif method == 'padding':
            max_height = max([arr.shape[0] for arr in images])
            max_width = max([arr.shape[1] for arr in images])
            padded = np.zeros(
                (len(images), max_height, max_width, 3),
                dtype=np.uint8
            )

            pads_per_image = []
            for idx, image in enumerate(images):
                diff_height = max(0, (max_height - image.shape[0]) / 2)
                diff_width = max(0, (max_width - image.shape[1]) / 2)
                pad_values = [
                    (
                        int(math.ceil(diff_height)),
                        int(math.floor(diff_height))
                    ),
                    (
                        int(math.ceil(diff_width)),
                        int(math.floor(diff_width))
                    ),
                    (0, 0),
                ]
                padded[idx, ...] = np.pad(image, pad_values)
                pads_per_image.append(pad_values)

            params['pads_per_image'] = pads_per_image
            return padded, params
        else:
            raise ValueError(
                'Invalid `method` set, options are `padding` or `crop`.'
            )

    def merge_out(poses_per_image, params):
        # If no merging occur, we have nothing to adjust.
        if not params['merged']:
            return poses_per_image

        if method == 'crop':
            raise NotImplementedError
        elif method == 'padding':
            new_poses_per_image = []
            for poses, pads in zip(poses_per_image, params['pads_per_image']):

                new_poses = []
                for pose in poses:
                    pads_per_axis = np.array([
                        pads[1][0], pads[0][0], 0
                    ]).reshape(1, -1)
                    keypoints = (
                        pose['keypoints'] - pads_per_axis
                    )
                    keypoints[keypoints[..., 2] == 0] = 0

                    new_poses.append({
                        'keypoints': keypoints,
                        'score': pose['score'],
                    })

                new_poses_per_image.append(new_poses)

            return new_poses_per_image
        else:
            raise ValueError(
                'Invalid `method` set, options are `padding` or `crop`.'
            )

    return merge_in, merge_out


class Estimation:

    def __init__(
        self, checkpoint=None, short_side=184, merge_method='padding',
        device=default_device, lazy=False
    ):
        """Initializes and loads the model for `checkpoint`.

        Parameters
        ----------
        checkpoint : str or None
            Checkpoint (and model) to use in order to perform pose estimation.
            If `None`, will use the default one for the task.
        short_side : int
            Resize images' short side to `short_side` before sending over to
            the pose estimation model. Default is `184` to keep the model fast
            enough, though for better results `386` is an appropriate value.
        merge_method : 'padding', 'crop'
            How to merge images together into a batch when receiving a
            list. Merge is done after resizing. Options are:

            * `padding`, which will add padding around images, possibly
              increasing total image size. If mixing portrait and landscape
              images, might be inefficient.

            * `crop`, which will center-crop the images to the smallest
              size. If images are of very different sizes, might end up
              cropping too much.
        device : torch.Device
            Device to load the model on.
        lazy : bool
            If set, will defer model loading until first call.

        """
        self.device = device
        self.estimation_cls = get_class_for_checkpoint(TASK_NAME, checkpoint)

        self.short_side = short_side

        # Load the model into memory unless we have the lazy loading set.
        self.model = (
            self.estimation_cls(
                device=self.device, short_side=self.short_side
            ) if not lazy else None
        )

        self.merge_in, self.merge_out = merge_factory(method=merge_method)

    def __repr__(self):
        return f'<Estimation({self.estimation_cls.__name__})>'

    def __call__(self, images):
        """Performs pose estimation on `images`.

        Derives the actual prediction to the model the `Estimation` object was
        initialized with.

        Parameters
        ----------
        images : list or tuple or np.ndarray

        Returns
        -------
        list
            List of dictionaries containing pose data for a single image, or a
            list of these entries thereof.

        """
        # If `images` is a single `np.ndarray`, turn into a list.
        expanded = False
        if (
            not (isinstance(images, list) or isinstance(images, tuple))
            and len(images.shape) == 3
        ):
            expanded = True
            images = np.expand_dims(images, 0)

        # Run generic image preprocessing, such as merging a list of images
        # into a single numpy array.
        images, merge_params = self.merge_in(images)

        # Call the actual model on the images. If we haven't loaded the model
        # yet due to lazy loading, load it.
        if self.model is None:
            self.model = self.estimation_cls(
                device=self.device, short_side=self.short_side
            )
        out = self.model.call(images)

        # Run the generic image postprocessing.
        out = self.merge_out(out, merge_params)

        return out[0] if expanded else out


pose_estimation = Estimation(lazy=True)
"""Default entry point to pose estimation.

This is an instantiation of the :class:`Estimation` class, lazily-loaded in
order to avoid reading the checkpoints on import. Refer to that class'
:func:`__call__ <Estimation.__call__>` method for more information.
"""



import ffmpeg
import os
import subprocess

from threading import Thread
from queue import Queue

from terran.io.video import DEFAULT_WRITER_BUFFER_SIZE, VideoClosed
from terran.io.video.reader import Video, open_video


def _frame_writer(queue, cmd):
    proc = subprocess.Popen(
        cmd,
        stdin=subprocess.PIPE,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL
    )

    while True:
        # TODO: Check for changing sizes in the _frame_writer func.
        frame_or_func = queue.get()
        if frame_or_func is None:
            break

        frame_or_func, *args = frame_or_func
        if hasattr(frame_or_func, '__call__'):
            frame = frame_or_func(*args)
        else:
            frame = frame_or_func

        proc.stdin.write(frame.tobytes())

    proc.stdin.close()

    # If process hasn't ended yet, terminate it, giving it some time to finish
    # up the file.
    if proc.poll() is None:
        proc.terminate()
        proc.wait(timeout=10.0)


class VideoWriter:

    def __init__(
        self, output_path, framerate=None, copy_format_from=None,
        size_hint=None, **kwargs
    ):
        """Initialize the writing of a video.

        Video format (codecs, framerate, etc.) can be specified using the
        keyword arguments, which will be fed to `ffmpeg` as-is, or by copying
        them from an existing `Video`.

        Parameters
        ----------
        framerate (int, str or None): Framerate of the output video. Will pass
            along to `ffmpeg`.  Can be an `int` with the frames per second, a
            `str` with a fraction (e.g. `'5000/1001'`). Default is `30`, or the
            framerate of `copy_format_from`, if specified.
        copy_format_from : str, pathlib.Path or Video
            Either `Video` instance or path to video to copy format from.
        size_hint : tuple
            (height, width) tuple indicating the frame size the video will
            have. Defaults to `None`, meaning that it will be infered from the
            first frame.

        """
        self.output_path = os.path.expanduser(output_path)

        # Calculate the framerate for the video. Priority is: value of
        # `framerate`, value from `copy_format_from`, or default of `30`.
        if framerate is None and copy_format_from is None:
            self.framerate = 30
        elif framerate is None and copy_format_from is not None:
            if not isinstance(copy_format_from, Video):
                # We got a path to a video, open it to check the framerate.
                copy_format_from = open_video(copy_format_from)
            self.framerate = copy_format_from.framerate
        else:
            self.framerate = framerate

        self.size_hint = size_hint

        # Handler for video-writing thread and queue.
        self._thread = None
        self._queue = None
        self._closed = False

    def __del__(self):
        if not self._closed:
            self.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def _prepare_ffmpeg_cmd(self):
        spec = ffmpeg

        kwargs = {
            'framerate': str(self.framerate),
            'format': 'rawvideo',
            'pix_fmt': 'rgb24',
            's': '{}x{}'.format(self.width, self.height),  # Frame size.
        }
        spec = spec.input('pipe:', **kwargs)

        kwargs = {
            'pix_fmt': 'yuv420p',
        }
        spec = spec.output(self.output_path, **kwargs)

        spec = spec.global_args('-y')

        spec = spec.compile()

        return spec

    def write_frame(self, frame_or_func, *args):
        """Write a frame or the result of a rendering function to the video.

        Note that if no `size_hint` is provided when creating the writer, and a
        rendering function is passed as argument, the rendering function might
        be executed twice, in order to infer the size of the video to write.

        """
        if self._closed:
            raise VideoClosed('The video has already been closed.')

        if not self._thread:
            # If no size hint is provided, get the dimensions of the video from
            # the first frame fed. If it's a function, we need to execute it
            # once.
            if not self.size_hint:
                if hasattr(frame_or_func, '__call__'):
                    frame = frame_or_func(*args)
                else:
                    frame = frame_or_func
                self.height, self.width = frame.shape[0:2]
            else:
                self.height, self.width = self.size_hint

            cmd = self._prepare_ffmpeg_cmd()

            self._queue = Queue(DEFAULT_WRITER_BUFFER_SIZE)
            # TODO: Daemon or not?
            self._thread = Thread(
                target=_frame_writer,
                args=(self._queue, cmd)
            )
            self._thread.start()

        self._queue.put((frame_or_func, *args))

    def close(self):
        if self._closed:
            raise VideoClosed('The video has already been closed.')

        if self._thread:
            self._queue.put(None)
            self._thread.join()
            self._closed = True


def write_video(*args, **kwargs):
    """Creates a writer for a video.

    Arguments are passed verbatim to `VideoWriter`.

    Returns
    -------
    VideoWriter
        `VideoWriter` instance representing the writer object that can be fed
        frames.

    """
    return VideoWriter(*args, **kwargs)



import numpy as np
import torch

from PIL import Image
from sklearn.preprocessing import normalize
from skimage.transform import SimilarityTransform

from terran import default_device
from terran.checkpoint import get_checkpoint_path
from terran.face.recognition.arcface.model import FaceResNet100


def load_model():
    model = FaceResNet100()
    model.load_state_dict(torch.load(
        get_checkpoint_path('terran.face.recognition.arcface.ArcFace')
    ))
    model.eval()
    return model


def preprocess_face(image, landmark, image_size=(112, 112)):
    """Prepares the face image for the recognition model.

    Uses the detected landmarks from the face detection stage, then aligns
    the image, pads to `image_side`x`image_side`, and turns it into the BGR
    CxHxW format.

    Parameters
    ----------
    image : np.ndarray of size HxWxC.
        Image containing a face to preprocess.
    landmark : np.ndarray of size (5, 2).
        Landmark coordinates.

    """

    # Target location of the facial landmarks.
    src = np.array([
        [30.2946, 51.6963],
        [65.5318, 51.5014],
        [48.0252, 71.7366],
        [33.5493, 92.3655],
        [62.7299, 92.2041]
    ], dtype=np.float32)

    if image_size[1] == 112:
        src[:, 0] += 8.0

    dst = landmark.astype(np.float32)

    t_form = SimilarityTransform()
    t_form.estimate(dst, src)

    #
    # Do align using landmark.
    #
    # The Image.transform method requires the inverted transformation matrix,
    # without the last row, and flattened
    #
    t_matrix = np.linalg.inv(t_form.params)[0: -1, :].flatten()

    warped = Image.fromarray(image).transform(
        size=(image_size[1], image_size[0]),
        method=Image.AFFINE,
        data=t_matrix,
        resample=Image.BILINEAR,
        fillcolor=0,
    )

    warped = np.array(warped)
    return warped.transpose([2, 0, 1])[::-1, ...]


def preprocess_face_no_landmarks(image, image_side=112):
    """Preprocess a face without landmarks.

    Resize image to have side `image_side` and add padding around it.
    """
    face = Image.fromarray(image)

    scale = image_side / max(face.size[0], face.size[1])
    face = face.resize(
        (int(face.size[0] * scale), int(face.size[1] * scale))
    )

    x_min = int((image_side - face.size[0]) / 2)
    x_max = int((image_side - face.size[0]) / 2) + face.size[0]
    y_min = int((image_side - face.size[1]) / 2)
    y_max = int((image_side - face.size[1]) / 2) + face.size[1]

    preprocessed = np.zeros(
        (3, image_side, image_side), dtype=np.uint8
    )
    preprocessed[:, y_min:y_max, x_min:x_max] = (
        np.asarray(face).transpose([2, 0, 1])[::-1, ...]
    )

    return preprocessed


class ArcFace:

    def __init__(self, device=default_device, image_side=112):
        self.device = device
        self.model = load_model().to(self.device)
        self.image_side = image_side

    def call(self, images, faces_per_image=None):
        """Performs feature extraction on `images`.

        Can either receive already-cropped images for faces (when
        `faces_per_image` is `None`), or a list of images and the face
        coordinates and landmarks to crop, as returned by a `Detection`
        function. The latter is preferable, as a more precise preprocessing can
        be performed.

        Parameters
        ----------
        images : list of numpy.ndarray or numpy.ndarray
            Images to perform face recognition on.
        faces_per_image : list of list of dicts.
            Each dict entry must contain `bbox` and `landmarks` keys, as
            returned by a `Detection` instance.

        Returns
        -------
        list of numpy.ndarray or numpy.ndarray
            One entry per image, with a numpy array of size (N_i, F), with F
            being the embedding size returned by the model.

        """
        # Since every image will have a small number of faces, we want to
        # unroll the features so we can use larger batch sizes with the model,
        # and then re-arrange them in the original structure again. For that,
        # we keep track of the indices with which to split in `splits`.
        preprocessed = []
        if faces_per_image is not None:
            for image, faces in zip(images, faces_per_image):
                for face in faces:
                    preprocessed.append(
                        preprocess_face(image, face['landmarks'])
                    )

            # A bit of magic, but gets us the index within `preprocessed` where
            # faces of each image start.
            splits = np.cumsum(list(map(len, faces_per_image)))[:-1]
        else:
            # No landmarks provided, so preprocess it manually.
            for image in images:
                preprocessed.append(
                    preprocess_face_no_landmarks(image, self.image_side)
                )

            # No splits to perform under this scenario, we pack them up as if
            # they were all faces for a single image.
            splits = []

        # No faces received, return early.
        if not preprocessed:
            # TODO: Embedding output depends on the model.
            return [
                np.empty((0, 512)) for _ in images
            ]

        preprocessed = np.stack(preprocessed, axis=0)

        # Now turn the (already preprocessed) input into a `torch.Tensor` and
        # feed through the network.
        data = torch.tensor(
            preprocessed, device=self.device, dtype=torch.float32
        )
        with torch.no_grad():
            features = self.model(data).cpu().numpy()

        features = normalize(features, axis=1)

        features_per_image = np.split(features, splits, axis=0)

        if faces_per_image is None:
            # If no per-image separation, just return the features.
            features_per_image = features_per_image[0]

        return features_per_image



import cairo
import math
import numpy as np

from cairo import Context, ImageSurface
from functools import wraps

from terran.pose import Keypoint
from terran.vis import (
    FACE_COLORMAP, MARKER_SCALES, POSE_CONNECTIONS, POSE_CONNECTION_COLORS,
    POSE_KEYPOINT_COLORS,
)


def with_cairo(vis_func):
    """Wrapper function to prepare the cairo context for the vis function."""

    @wraps(vis_func)
    def func(image, objects, *args, **kwargs):
        # Allow sending in a single object in every function.
        if not (isinstance(objects, list) or isinstance(objects, tuple)):
            objects = [objects]

        # Calculate the appropriate scaling for the markers.
        area = image.shape[1] * image.shape[0]
        for ref_area, scale in MARKER_SCALES:
            if area >= ref_area:
                break

        # TODO: Ideally, we would like to avoid having to create a copy of the
        # array just to add paddings to support the cairo format, but there's
        # no way to use a cairo surface with 24bpp.

        # TODO: Take into account endianness of the machine, as it's possible
        # it has to be concatenated in the other order in some machines.

        # We need to add an extra `alpha` layer, as cairo only supports 32 bits
        # per pixel formats, and our numpy array uses 24. This means creating
        # one copy before modifying and a second copy afterwards.
        with_alpha = np.concatenate(
            [
                image[..., ::-1],
                255 * np.ones(
                    (image.shape[0], image.shape[1], 1),
                    dtype=np.uint8
                )
            ], axis=2
        )

        surface = ImageSurface.create_for_data(
            with_alpha,
            cairo.Format.RGB24,
            image.shape[1],
            image.shape[0]
        )

        ctx = Context(surface)

        # Set up the font. If not available, will default to a Sans Serif
        # font available in the system, so no need to have fallbacks.
        ctx.select_font_face(
            "DejaVuSans-Bold",
            cairo.FONT_SLANT_NORMAL,
            cairo.FONT_WEIGHT_NORMAL
        )
        ctx.set_font_size(int(16 * scale))

        vis_func(ctx, objects, scale=scale, *args, **kwargs)

        # Return the newly-drawn image, excluding the extra alpha channel
        # added.
        image = with_alpha[..., :-1][..., ::-1]

        return image

    return func


def draw_marker(ctx, coords, color=(255, 0, 0), scale=1):
    """Draw a marker on `ctx` at `coords`.

    The marker itself is a rectangle with rounded corners.
    """
    x_min, y_min, x_max, y_max = coords
    width = x_max - x_min
    height = y_max - y_min

    degrees = math.pi / 180.0

    radius = 10.0 * scale
    ctx.set_source_rgba(*color, 1.0)
    ctx.set_line_width(3. * scale)
    ctx.set_dash([])

    ctx.new_sub_path()
    ctx.arc(
        x_min + width - radius, y_min + radius,
        radius, -90 * degrees, 0 * degrees
    )
    ctx.arc(
        x_min + width - radius, y_min + height - radius,
        radius, 0 * degrees, 90 * degrees
    )
    ctx.arc(
        x_min + radius, y_min + height - radius,
        radius, 90 * degrees, 180 * degrees
    )
    ctx.arc(
        x_min + radius, y_min + radius,
        radius, 180 * degrees, 270 * degrees
    )
    ctx.close_path()

    ctx.stroke()

    ctx.set_dash([10. * scale])
    ctx.set_line_width(1. * scale)

    ctx.move_to((x_min + x_max) / 2, y_min)
    ctx.line_to((x_min + x_max) / 2, y_max)

    ctx.move_to(x_min, (y_min + y_max) / 2)
    ctx.line_to(x_max, (y_min + y_max) / 2)

    ctx.stroke()


@with_cairo
def vis_faces(ctx, faces, scale=1.0):
    """Draw boxes over the detected faces for the given image.

    Parameters
    ----------
    image : np.ndarray representing an image
        Image to draw faces over.
    faces : dict or list of dicts (from `face_detection` or `face_tracking`)
        Faces to draw on `image`. The expected format is the one returned from
        `face_detection` or `face_tracking`, with two optional extra fields:

        - ``text`` (str): Text to be written next to the box.
        - ``name`` (str): Name associated to the face, in order to make the
          color used for the box fixed.

        If available, the ``track`` field will be used as the default value
        for the two values above, if they aren't specified.

    Returns
    -------
    np.ndarray
        Copy of `image` with the faces drawn over.

    """
    for face in faces:
        # Get the name and text for the current face.
        face_name = face.get('name') or face.get('track')
        if face.get('text') is not None:
            face_text = face['text']
        elif face.get('track') is not None:
            face_text = f"#{face['track']}"
        else:
            face_text = None

        color = map(lambda x: x / 255, FACE_COLORMAP(face_name))
        draw_marker(ctx, face['bbox'], color=color, scale=scale)

        if face_text is not None:
            ctx.move_to(
                face['bbox'][0] + 3 * scale,
                face['bbox'][1] + 15 * scale
            )
            ctx.show_text(face_text)


def draw_keypoints(ctx, keypoints, scale=1.0):
    for keypoint in keypoints:
        for idx, (x, y, is_present) in enumerate(keypoint['keypoints']):
            if not is_present:
                continue

            color = map(
                lambda x: x / 255,
                POSE_KEYPOINT_COLORS[Keypoint(idx)]
            )
            ctx.set_source_rgba(*color, 0.9)

            ctx.arc(x, y, 3 * scale, 0, 2 * math.pi)
            ctx.fill()
            ctx.stroke()


def draw_limbs(ctx, keypoints, scale=1.0):
    for keypoint in keypoints:
        kps = keypoint['keypoints']
        for idx, (conn_src, conn_dst) in enumerate(POSE_CONNECTIONS):
            x_src, y_src, src_present = kps[conn_src.value]
            x_dst, y_dst, dst_present = kps[conn_dst.value]

            # Ignore limbs for which one of the keypoints is missing.
            if not (src_present and dst_present):
                continue

            color = map(lambda x: x / 255, POSE_CONNECTION_COLORS[idx])
            ctx.set_source_rgba(*color, 0.7)
            ctx.set_line_width(1.)

            # We'll use a Bezier curve using a rectangle around the line as
            # control points, so we first calculate the normal and the
            # direction we need to move the points on in order to have a
            # constant-height box around the lines.
            width = 4 * scale

            if abs(y_dst - y_src) > 0:
                normal = - (x_dst - x_src) / (y_dst - y_src)
                x_base = width / math.sqrt(normal ** 2 + 1)
                y_base = x_base * normal
            else:
                # Careful calculating normal if limb is horizontal.
                x_base = 0
                y_base = width

            ctx.move_to(x_src, y_src)
            ctx.curve_to(
                int(x_src + x_base), int(y_src + y_base),
                int(x_dst + x_base), int(y_dst + y_base),
                x_dst, y_dst,
            )
            ctx.curve_to(
                int(x_dst - x_base), int(y_dst - y_base),
                int(x_src - x_base), int(y_src - y_base),
                x_src, y_src,
            )
            ctx.fill()

            ctx.stroke()


@with_cairo
def vis_poses(ctx, poses, scale=1.0):
    """Draw boxes over the detected poses for the given image.

    Parameters
    ----------
    image : np.ndarray representing an image.
        Image to draw faces over.
    poses : dict or list of dicts, as returned by `pose_estimation`
        Poses to draw on `image`. The expected format is the one returned from
        `pose_estimation`.

    Returns
    -------
    np.ndarray
        Copy of `image` with the poses drawn over.

    """
    draw_limbs(ctx, poses, scale=scale)
    draw_keypoints(ctx, poses, scale=scale)



import ffmpeg
import json
import math
import numpy as np
import os
import subprocess

from threading import Event, Thread
from queue import Full as QueueFull, Queue

from terran.io.video import DEFAULT_READER_BUFFER_SIZE, EndOfVideo, VideoClosed


def youtube_dl_available():
    """Check if `youtube-dl` is available in the installation."""
    try:
        import youtube_dl  # noqa
        return True
    except ImportError:
        return False


def ffmpeg_probe(path, **kwargs):
    """Run ffprobe on the specified file and return a JSON representation of
    the output.

    Based on the `ffmpeg.probe` provided by `ffmpeg-python`, but allows passing
    along additional options to `ffmpeg`.

    Parameters
    ----------
    path : str
        This parameter can be a path or URL pointing directly to a video file
        or stream.

    Raises
    ------
    ffmpeg.Error
        If `ffprobe` returns a non-zero exit code. The stderr output can be
        retrieved by accessing the `stderr` property of the exception.

    """
    if not is_path_stream(path):
        path = os.path.expanduser(path)

    # Gather all `kwargs` into a list of tokens to send to `ffprobe`.
    additional_args = []
    for key, value in kwargs.items():
        if not key.startswith('-'):
            key = f'-{key}'
        additional_args.extend([key, str(value)])

    args = [
        'ffprobe', *additional_args, '-show_format', '-show_streams', '-of',
        'json', path
    ]

    proc = subprocess.Popen(
        args, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )

    out, err = proc.communicate()
    if proc.returncode != 0:
        raise ffmpeg.Error('ffprobe', out, err)

    return json.loads(out.decode('utf-8'))


def is_path_stream(path):
    """Returns True if the path points to a video stream."""
    return any([
        path.startswith(prefix)
        for prefix in ['/dev/', 'http://', 'https://']
    ])


def parse_timestamp(timestamp):
    if '.' in timestamp:
        timestamp, miliseconds = timestamp.split('.')
        miliseconds = float('0.{}'.format(miliseconds))
    else:
        miliseconds = 0.0
    hours, minutes, seconds = map(float, timestamp.split(':'))
    time = hours * 60 * 60 + minutes * 60 + seconds + miliseconds
    return time


def _read_batch_from_stream(proc, spec):
    width = spec['width']
    height = spec['height']
    batch_size = spec['batch_size']

    # `rgb24` uses three bytes per pixel.
    bytes_to_read = width * height * 3 * (
        batch_size if batch_size is not None else 1
    )
    buffer = proc.stdout.read(bytes_to_read)
    if len(buffer) == 0:
        # No data read, we reached the end of the video.
        return
    elif len(buffer) < bytes_to_read:
        # Received less bytes than expected, must've been unable to
        # complete a batch. Calculate the number of frames read.
        frames_read = int(len(buffer) / width / height / 3)
    else:
        # Received the expected amount data, continue onwards.
        frames_read = batch_size

    frames = np.frombuffer(buffer, np.uint8)
    if batch_size is not None:
        frames = frames.reshape(
            [frames_read, height, width, 3]
        )
    else:
        frames = frames.reshape([height, width, 3])

    return frames


def _clean_up_proc(proc):
    if proc.poll() is None:
        # It's a reader process, we don't care too much for it to end cleanly.
        proc.kill()


def _frame_reader(queue, should_stop, cmd, spec):
    """Worker function for the reading thread."""
    # Open the `ffmpeg` subprocess.
    proc = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL,
    )

    while proc.poll() is None:
        frames = _read_batch_from_stream(proc, spec)
        if frames is None:
            break

        # Since there's no way to wait on either a `queue.put()` or a stop
        # signal, we must work around it by using a timeout on the
        # `queue.put()` function and periodically checking for the stop signal.
        while True:
            # Stop signal set, clean up the underlying process and return.
            if should_stop.is_set():
                _clean_up_proc(proc)
                return

            # Attempt to place batch in frame, blocking up to one second. If
            # queue is full, try again; if not, break the loop and continue
            # with the next batch.
            try:
                queue.put(frames, timeout=1.0)
            except QueueFull:
                continue

            break

    # After all frames are read, add a sentinel to the queue and terminate the
    # process (if it hasn't already).
    queue.put(None)
    _clean_up_proc(proc)


class Video:
    """Container class for a video, for interfacing with an underlying input.

    Is tasked with handling the file descriptiors, with buffering the stream,
    with dropping frames when needed, with decoding frames out of the video,
    and more.
    """

    def __init__(
        self, path, batch_size=None, framerate=None, is_stream=None,
        read_for=None, start_time=None, ydl_format='best'
    ):
        """Initializes the reading of the video, performing the necessary
        checks.

        Note that if `duration` is `None` and `path` points to a stream,
        `Video` will return frames indefinitely.

        Parameters
        ----------
        path : str
            Path to a video file, stream URL or capture device.
        batch_size : int or None
            Batch size for the returned frames. If `None`, no batching will
            occur (thus, a rank 3 array will be returned).
        framerate : int, str or None
            Framerate to output frames as. Will pass along to `ffmpeg`, which
            will drop or add frames as needed.  Can be an `int` with the frames
            per second, a `str` with a fraction (e.g. `'5000/1001'`) or `None`
            to leave as-is.
        is_stream : bool or None
            Whether the video is a capture device, a stream or somehow a
            never-ending video file. If `None`, will try to guess.
        read_for : int, float or None
            Maximum number of seconds to read, or `None` for reading until the
            end.

            If `read_for` is bigger than the source duration (considering
            the start time), the actual video duration might be less than
            this value.
        start_time : int, str or None
            Time to start video from. If an `int`, specified in seconds. If
            `str`, specified as a timestamp with the format `HH:MM:SS.ms`.
        ydl_format : str
            The format filtering option for YouTube-DL. Check out the
            YouTube-DL "Format Selection" documentation for more information:
            https://github.com/ytdl-org/youtube-dl#format-selection

        """
        self.path = os.path.expanduser(path)

        self.batch_size = batch_size
        self.read_for = read_for
        self._framerate = framerate
        self.ydl_format = ydl_format

        # Parse before storing, so `start_time` will be either an `int`
        # specifying the seconds or `None`.
        if isinstance(start_time, str):
            start_time = parse_timestamp(start_time)
        self.start_time = start_time

        # Try to guess if it's capture device or a stream, as we might need to
        # do things differently.
        if not is_stream:
            self.is_stream = is_path_stream(self.path)

        # Check for existing video streams within the file first, to validate
        # everything and to get the video dimensions.
        try:
            if self.is_stream:
                self.stream_path = self._get_stream_path()

                # If a capture device, use a bigger `probesize` and
                # `analyzeduration` to make sure we have information on all
                # existing streams.
                # TODO: Use a lower probesize and analyzeduration and keep
                # increasing on retries.
                probe = ffmpeg_probe(
                    self.stream_path,
                    probesize=20 * 1024 * 1024,
                    analyzeduration=10 * 1000 * 1000,
                )
            else:
                probe = ffmpeg_probe(self.path)
        except ffmpeg.Error:
            message = f'Video at `{path}` not found. Are you sure it exists?'
            if not youtube_dl_available():
                message += (
                    "\n\n"
                    "Unable to find suitable way to stream from online video "
                    "platforms. If you're trying to stream from YouTube or "
                    "other streaming platforms, make sure `youtube-dl` is "
                    "installed first. If not, ignore this message."
                )

            raise ValueError(message)

        video_stream = next(
            (
                stream for stream in probe['streams']
                if stream.get('codec_type') == 'video'
            ), None
        )
        if not video_stream:
            raise ValueError(
                f'No video stream found at `{path}`. Are you sure this is a '
                'video file or stream?'
            )

        # Extract the information we need from the underlying source video.
        self.width = int(video_stream['width'])
        self.height = int(video_stream['height'])

        # Get the framerate of the video source. Uses `avg_frame_rate`, as
        # `r_frame_rate` will default to the lowest common denominator when
        # multiple streams are present (such as on DVB devices). Should be
        # accurate enough.
        if '/' in video_stream['avg_frame_rate']:
            num, den = map(int, video_stream['avg_frame_rate'].split('/'))
            self.source_framerate = num / den
        else:
            self.source_framerate = float(video_stream['avg_frame_rate'])

        # Attempt to get the duration either from the selected video stream or
        # from the container data.
        self.source_duration = None
        if 'duration' in video_stream:
            self.source_duration = float(video_stream['duration'])
        elif 'duration' in probe.get('format', {}):
            self.source_duration = float(probe['format']['duration'])

        if self.duration is not None and self.duration < 0:
            raise ValueError(
                'Duration of the video is negative. Is the `start_time` '
                'timestamp after the video ends?'
            )

        # Handler for video-reading thread and queue.
        self._thread = None
        self._queue = None
        self._closed = False

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def __iter__(self):
        return self

    def __next__(self):
        try:
            frames = self.read_frames()
        except EndOfVideo:
            raise StopIteration
        return frames

    def __del__(self):
        if not self._closed:
            self.close()

    def __len__(self):
        """Returns number of batches present in the video.

        Raises
        ------
        AttributeError
            If the video is a stream or we were otherwise unable to infer the
            video length.

        """
        if not self.duration:
            raise AttributeError(
                'Video doesn\'t have a duration. Is it a stream?'
            )

        batch_size = self.batch_size if self.batch_size else 1
        return math.ceil(
            math.ceil(self.duration * self.framerate) / batch_size
        )

    @property
    def framerate(self):
        """Effective framerate of the video output.

        Either the requested framerate, if set, or the source's.
        """
        return (
            self._framerate if self._framerate is not None
            else self.source_framerate
        )

    @property
    def duration(self):
        """Effective duration of the video output, in seconds.

        Considers the `read_for` attribute, the duration of the source video,
        and the `start_time` set by the user.

        Returns
        -------
        float or None
            Time in seconds or `None` if it's a stream.

        """
        if not self.source_duration:
            # If no duration, will read for `self.read_for` seconds, which may
            # be `None`.
            return self.read_for

        # If source duration is available, consider the time that was skipped
        # due to `self.start_time`.
        source_duration = (
            self.source_duration if not self.start_time
            else self.source_duration - self.start_time
        )
        if self.read_for:
            return min(source_duration, self.read_for)
        else:
            return source_duration

    def _get_stream_path(self):
        """Check if video stream comes from a video sharing platform"""
        if not youtube_dl_available():
            return self.path

        import youtube_dl

        ydl_options = {
            'format': self.ydl_format,
            'quiet': True,
            'no_warnings': True
        }

        for extractor in youtube_dl.gen_extractors():
            if extractor.suitable(self.path):
                try:
                    with youtube_dl.YoutubeDL(ydl_options) as ydl:
                        stream_info = ydl.extract_info(
                            self.path, download=False
                        )
                        self.ydl_info = stream_info

                        if stream_info['url'] is None:
                            raise ValueError(
                                'Unable to find stream URL for video format '
                                f'{self.ydl_format}'
                            )
                        return stream_info['url']
                except youtube_dl.utils.YoutubeDLError:
                    break

        return self.path

    def _prepare_ffmpeg_cmd(self):
        """Prepare the subprocess command for ffmpeg."""
        spec = ffmpeg

        # Prepare the input parameters.
        kwargs = {
            'err_detect': 'ignore_err',
        }

        if self.duration is not None:
            kwargs.update({'t': str(self.duration)})

        if self.is_stream:
            # If a capture device, use a bigger `probesize` and
            # `analyzeduration` to make sure we have information on all
            # existing streams.
            kwargs.update({
                'probesize': 20 * 1024 * 1024,
                'analyzeduration': 10 * 1000 * 1000,
            })

        # Start time works for normal videos and YouTube streams
        if self.start_time or self.is_stream:
            # Seek after 5 seconds to allow ffmpeg to see a reference frame.
            # TODO: Why is this even necessary? Can't I just tell it to start
            # from where valid?
            kwargs.update({'ss': self.start_time or '00:00:05'})

        # Send the appropriate path to `ffmpeg`, either stream or normal.
        input_file = self.stream_path if self.is_stream else self.path
        spec = spec.input(input_file, **kwargs)

        # Prepare the output parameters.
        kwargs = {}

        # Only pass the flag if a specific framerate was requested by the user.
        if self._framerate:
            kwargs.update({'r': str(self._framerate)})

        spec = spec.output(
            'pipe:', format='rawvideo', pix_fmt='rgb24', **kwargs
        )

        cmd = spec.compile()
        return cmd

    def read_frames(self):
        # If the video resource has been opened and closed already, bail out.
        if self._closed:
            raise EndOfVideo

        # Open and manage the `ffmpeg` process explicitly for better
        # control. The process will be opened in a `Thread` (as it's mostly
        # I/O-bound and a thread provides a lower overhead) and will prefetch
        # `DEFAULT_READER_BUFFER_SIZE` batches in advance.
        if not self._thread:
            cmd = self._prepare_ffmpeg_cmd()
            spec = {
                'width': self.width,
                'height': self.height,
                'batch_size': self.batch_size,
            }

            self._queue = Queue(DEFAULT_READER_BUFFER_SIZE)
            self._stop_signal = Event()

            self._thread = Thread(
                args=(self._queue, self._stop_signal, cmd, spec),
                name='FrameReader',
                target=_frame_reader,
                # We don't want to keep the program hanging; as it's a reader
                # thread, we don't care that the resources are closed abruptly.
                daemon=True,
            )
            self._thread.start()

        frames = self._queue.get()
        if frames is None:
            raise EndOfVideo

        return frames

    def close(self):
        if self._closed:
            raise VideoClosed('The video has already been closed.')

        if self._thread:
            # Set the stop signal, making the thread stop trying to put frames
            # into the queue. Due to the hackish approach, it might take up to
            # one second to finish.
            self._stop_signal.set()
            self._thread.join()
            self._closed = True


def open_video(*args, **kwargs):
    """Opens a video file or stream.

    The function will guess whether the input is a video file, a capture device
    or a stream and will proceed as needed.

    Arguments are passed verbatim to `Video`.

    Returns
    -------
    Video
        `Video` instance representing the video at `path`.

    """
    return Video(*args, **kwargs)



# flake8: noqa
from terran.defaults import default_device

from terran.face import extract_features, face_detection
from terran.pose import pose_estimation

from terran.io import open_image, open_video

from terran.vis import display_image, vis_faces, vis_poses



import os

from setuptools import setup, find_packages


# Load the README file to use it as long description.
with open('README.md', 'r') as f:
    readme = f.read()


INSTALL_REQUIRES = [
    'click',
    'ffmpeg-python',
    'filterpy',
    'opencv-python-headless',
    'numpy',
    'Pillow',
    'requests',
    'scikit-image',
    'scikit-learn>=0.21',
    'torch',
    'torchvision',
]

# Dependencies that are to be removed if the environment is `READTHEDOCS`, so
# we don't run into troubles trying to install unsupported C-extensions. The
# list should match `autodoc_mock_imports` in `docs/conf.py`, but with the
# package names instead of the top-level modules.
MOCKED_IN_DOCS = [
    'opencv-python',
]

if os.environ.get('READTHEDOCS'):
    INSTALL_REQUIRES = list(
        set(INSTALL_REQUIRES) - set(MOCKED_IN_DOCS)
    )

setup(
    name='terran',
    version='0.1.2',
    author='Agustín Azzinnari',
    author_email='me@nagitsu.com',
    description='A human perception library',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://terran-project.com',

    packages=find_packages(),
    install_requires=INSTALL_REQUIRES,
    extras_requires={
        'cairo': ['pycairo'],
    },

    entry_points="""
        [console_scripts]
        terran=terran.cli:cli
    """,

    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Libraries',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ]
)



import numpy as np
import random

from PIL import Image
from io import BytesIO
from subprocess import run, SubprocessError

from terran.pose import Keypoint


MARKER_SCALES = [
    (1920 * 1080, 1.8),
    (1280 * 720, 1.5),
    (480 * 360, 1.3),
    (0, 1),
]


def display_image(image):
    """Displays an image using an external viewer.

    Will first try using `feh`, if it's installed locally, and `matplotlib` as
    fallback, if an error occurs.

    Parameters
    ----------
    image : np.array or PIL.Image
        Image to be displayed.

    Raises
    ------
    Exception
        If no suitable backend is found.

    """
    # If a numpy array, turn into a Pillow image first.
    if isinstance(image, np.ndarray):
        image = Image.fromarray(image)

    # Save the image compressed in a `BytesIO`.
    buf = BytesIO()
    image.save(buf, format='png')
    buf.seek(0)

    # Run the `feh` command, passing the buffer as input. If any error occurs,
    # fallback to using `imshow` from `matplotlib`.
    try:
        run(['feh', '-'], input=buf.read())
    except (FileNotFoundError, SubprocessError):
        try:
            import matplotlib.pyplot as plt
            plt.imshow(image)
            plt.show()
        except ImportError:
            raise Exception(
                'Unable to find a suitable backend to display an image. '
                'Tried `feh` and `matplotlib`. Install either in order to use '
                'this function.'
            )


def hex_to_rgb(x):
    """Turns a color hex representation into a tuple representation."""
    return tuple([int(x[i:i + 2], 16) for i in (0, 2, 4)])


def build_colormap():
    """Builds a colormap function that maps labels to colors.

    Returns:
        Function that receives a label and returns a color tuple `(R, G, B)`
        for said label.
    """
    # Build the 10-color palette to be used for all classes. The following are
    # the hex-codes for said colors (taken the default 10-categorical d3 color
    # palette).
    palette = (
        '1f77b4ff7f0e2ca02cd627289467bd8c564be377c27f7f7fbcbd2217becf'
    )
    colors = [hex_to_rgb(palette[i:i + 6]) for i in range(0, len(palette), 6)]

    seen_labels = {}

    def colormap(label=None):
        # If `label` is `None` return a random color, so we can have both fix
        # the colors and have them dynamic.
        if label is None:
            return random.choice(colors)

        # If label not yet seen, get the next value in the palette sequence.
        if label not in seen_labels:
            seen_labels[label] = colors[len(seen_labels) % len(colors)]

        return seen_labels[label]

    return colormap


FACE_COLORMAP = build_colormap()


POSE_CONNECTIONS = [
    (Keypoint.NOSE, Keypoint.NECK),
    (Keypoint.NOSE, Keypoint.R_EYE), (Keypoint.R_EYE, Keypoint.R_EAR),
    (Keypoint.NOSE, Keypoint.L_EYE), (Keypoint.L_EYE, Keypoint.L_EAR),

    (Keypoint.NECK, Keypoint.R_SHOULDER),
    (Keypoint.R_SHOULDER, Keypoint.R_ELBOW),
    (Keypoint.R_ELBOW, Keypoint.R_HAND),

    (Keypoint.NECK, Keypoint.R_HIP),
    (Keypoint.R_HIP, Keypoint.R_KNEE),
    (Keypoint.R_KNEE, Keypoint.R_FOOT),

    (Keypoint.NECK, Keypoint.L_SHOULDER),
    (Keypoint.L_SHOULDER, Keypoint.L_ELBOW),
    (Keypoint.L_ELBOW, Keypoint.L_HAND),

    (Keypoint.NECK, Keypoint.L_HIP),
    (Keypoint.L_HIP, Keypoint.L_KNEE),
    (Keypoint.L_KNEE, Keypoint.L_FOOT),
]


POSE_CONNECTION_COLORS = list(map(hex_to_rgb, [
    # Head.
    'e6550d', 'fd8d3c', 'fdae6b', '843c39', 'ad494a',

    # Right side.
    '637939', '8ca252', 'b5cf6b',
    '843c39', 'ad494a', 'd6616b',

    # Left side.
    '3182bd', '6baed6', '9ecae1',
    '8c6d31', 'bd9e39', 'e7ba52',
]))


POSE_KEYPOINT_COLORS = {
    Keypoint.NOSE: hex_to_rgb('e6550d'),
    Keypoint.NECK: hex_to_rgb('fd8d3c'),
    Keypoint.R_EYE: hex_to_rgb('fdae6b'),
    Keypoint.L_EYE: hex_to_rgb('843c39'),
    Keypoint.R_EAR: hex_to_rgb('ad494a'),
    Keypoint.L_EAR: hex_to_rgb('d6616b'),

    Keypoint.R_SHOULDER: hex_to_rgb('637939'),
    Keypoint.R_ELBOW: hex_to_rgb('8ca252'),
    Keypoint.R_HAND: hex_to_rgb('b5cf6b'),
    Keypoint.R_HIP: hex_to_rgb('843c39'),
    Keypoint.R_KNEE: hex_to_rgb('ad494a'),
    Keypoint.R_FOOT: hex_to_rgb('d6616b'),

    Keypoint.L_SHOULDER: hex_to_rgb('3182bd'),
    Keypoint.L_ELBOW: hex_to_rgb('6baed6'),
    Keypoint.L_HAND: hex_to_rgb('9ecae1'),
    Keypoint.L_HIP: hex_to_rgb('8c6d31'),
    Keypoint.L_KNEE: hex_to_rgb('bd9e39'),
    Keypoint.L_FOOT: hex_to_rgb('e7ba52'),
}


try:
    from terran.vis.cairo import (  # noqa
        vis_poses, vis_faces,
    )
except ImportError:
    from terran.vis.pillow import (  # noqa
        vis_poses, vis_faces,
      )



import torch
import torch.nn as nn
import torch.nn.functional as F


class ConvSepBlock(nn.Module):

    def __init__(self, in_c, out_c, stride=1, return_both=False):
        """Building block for base network.

        Consists of common Conv, BN and ReLU sequence, followed by the same
        sequence but with a separable Conv.

        Paramters
        ---------
        return_both : bool
            Return the outputs of both inner components, the conv and the
            separable blocks. We do this because it's the conv block the one
            that's used as feature pyramid.

        """
        super().__init__()

        self.return_both = return_both

        self.conv_block = nn.Sequential(
            nn.Conv2d(in_c, out_c, 1, stride=1, bias=False),
            nn.BatchNorm2d(out_c, momentum=0.9, eps=1e-5),
            nn.ReLU(),
        )

        self.sep_block = nn.Sequential(
            nn.Conv2d(
                out_c, out_c, 3, stride=stride, padding=1, groups=out_c,
                bias=False
            ),
            nn.BatchNorm2d(out_c, momentum=0.9, eps=1e-5),
            nn.ReLU(),
        )

    def forward(self, x):
        conv = self.conv_block(x)
        sep = self.sep_block(conv)

        if self.return_both:
            out = conv, sep
        else:
            out = sep

        return out


class BaseNetwork(nn.Module):

    def __init__(self):
        super().__init__()

        # The first convolutional block is different from the rest: the large
        # stride and kernel size is on the non-separable convolution instead.
        self.first_conv_block = nn.Sequential(
            nn.Conv2d(3, 8, 3, stride=2, padding=1, bias=False),
            nn.BatchNorm2d(8, momentum=0.9, eps=1e-5),
            nn.ReLU(),

            nn.Conv2d(8, 8, 3, stride=1, padding=1, groups=8, bias=False),
            nn.BatchNorm2d(8, momentum=0.9, eps=1e-5),
            nn.ReLU(),
        )

        # We're going to extract intermediate feature maps at downsampling
        # ratios of 8 and 16, so we group the outputs in a way that we can
        # retrieve them easily enough on the `forward` pass.
        self.scales = nn.ModuleList([
            nn.Sequential(
                ConvSepBlock(8, 16, stride=2),

                ConvSepBlock(16, 32),
                ConvSepBlock(32, 32, stride=2),

                ConvSepBlock(32, 64),
                ConvSepBlock(64, 64, stride=2, return_both=True),
            ),

            nn.Sequential(
                ConvSepBlock(64, 128),
                ConvSepBlock(128, 128),
                ConvSepBlock(128, 128),
                ConvSepBlock(128, 128),
                ConvSepBlock(128, 128),
                ConvSepBlock(128, 128, stride=2, return_both=True),
            ),
        ])

        self.final_conv = nn.Sequential(
            ConvSepBlock(128, 256),
            nn.Conv2d(256, 256, 1, bias=False),
            nn.BatchNorm2d(256, momentum=0.9, eps=1e-5),
            nn.ReLU(),
        )

    def forward(self, x):
        out = self.first_conv_block(x)

        feature_maps = []
        for scale in self.scales:
            conv, out = scale(out)
            feature_maps.append(conv)

        out = self.final_conv(out)
        feature_maps.append(out)

        return feature_maps


class ContextModule(nn.Module):
    """Context module to expand the receptive field of the feature map.

    Every point in the feature map will be a mixture of a 3x3, a 5x5 and a 7x7
    receptive fields. The first 128 channels correspond to the 3x3 one, and the
    remaining 64 and 64 to the rest.
    """

    def __init__(self):
        super().__init__()

        self.context_3x3 = nn.Sequential(
            nn.Conv2d(64, 32, 3, padding=1),
            nn.BatchNorm2d(32, momentum=0.9, eps=2e-5),
            nn.ReLU(),
        )

        self.dimension_reducer = nn.Sequential(
            nn.Conv2d(64, 16, 3, padding=1),
            nn.BatchNorm2d(16, momentum=0.9, eps=2e-5),
            nn.ReLU(),
        )

        self.context_5x5 = nn.Sequential(
            nn.Conv2d(16, 16, 3, padding=1),
            nn.BatchNorm2d(16, momentum=0.9, eps=2e-5),
            nn.ReLU(),
        )

        self.context_7x7 = nn.Sequential(
            nn.Conv2d(16, 16, 3, padding=1),
            nn.BatchNorm2d(16, momentum=0.9, eps=2e-5),
            nn.ReLU(),

            nn.Conv2d(16, 16, 3, padding=1),
            nn.BatchNorm2d(16, momentum=0.9, eps=2e-5),
            nn.ReLU(),
        )

    def forward(self, x):
        red = self.dimension_reducer(x)

        # By applying successive 3x3 convolutions, we get an effective
        # receptive field of 5x5 and 7x7.
        ctx_3x3 = self.context_3x3(x)
        ctx_5x5 = self.context_5x5(red)
        ctx_7x7 = self.context_7x7(red)

        out = torch.cat([ctx_3x3, ctx_5x5, ctx_7x7], dim=1)

        return out


class PyramidRefiner(nn.Module):
    """Refines the feature pyramids from the base network into usable form.

    Normalizes channel sizes, mixes them up, and runs them through the context
    module.
    """

    def __init__(self):
        super().__init__()

        self.conv_stride8 = nn.Sequential(
            nn.Conv2d(64, 64, 1),
            nn.BatchNorm2d(64, momentum=0.9, eps=2e-5),
            nn.ReLU(),
        )
        self.conv_stride16 = nn.Sequential(
            nn.Conv2d(128, 64, 1),
            nn.BatchNorm2d(64, momentum=0.9, eps=2e-5),
            nn.ReLU(),
        )
        self.conv_stride32 = nn.Sequential(
            nn.Conv2d(256, 64, 1),
            nn.BatchNorm2d(64, momentum=0.9, eps=2e-5),
            nn.ReLU(),
        )

        self.aggr_stride8 = nn.Sequential(
            nn.Conv2d(64, 64, 3, padding=1),
            nn.BatchNorm2d(64, momentum=0.9, eps=2e-5),
            nn.ReLU(),
        )
        self.aggr_stride16 = nn.Sequential(
            nn.Conv2d(64, 64, 3, padding=1),
            nn.BatchNorm2d(64, momentum=0.9, eps=2e-5),
            nn.ReLU(),
        )

        self.context_stride8 = ContextModule()
        self.context_stride16 = ContextModule()
        self.context_stride32 = ContextModule()

    def forward(self, x):
        """Forward pass for refiner.

        Expects `x` to be an array of three tensors, one per feature. Returns
        the same.
        """
        stride8, stride16, stride32 = x

        # First run every scale through the conv layers.
        proc_stride8 = self.conv_stride8(stride8)
        proc_stride16 = self.conv_stride16(stride16)
        proc_stride32 = self.conv_stride32(stride32)

        # Upsample the smaller to the bigger ones and sum. When upsampling,
        # resulting tensor might be bigger by one spatial dimension (if initial
        # dimension wasn't power of 2), so we slice the tensor to the same
        # size.
        ups_stride32 = F.interpolate(
            proc_stride32, scale_factor=2
        )[:, :, :proc_stride16.shape[2], :proc_stride16.shape[3]]
        proc_stride16 = self.aggr_stride16(
            proc_stride16 + ups_stride32
        )

        ups_stride16 = F.interpolate(
            proc_stride16, scale_factor=2
        )[:, :, :proc_stride8.shape[2], :proc_stride8.shape[3]]
        proc_stride8 = self.aggr_stride8(
            proc_stride8 + ups_stride16
        )

        # Now run every scale through a context module.
        ctx_stride8 = self.context_stride8(proc_stride8)
        ctx_stride16 = self.context_stride16(proc_stride16)
        ctx_stride32 = self.context_stride32(proc_stride32)

        return [ctx_stride8, ctx_stride16, ctx_stride32]


class OutputsPredictor(nn.Module):
    """Uses the feature pyramid to predict the final deltas for the network."""

    def __init__(self):
        super().__init__()

        # There are two anchors per point: two scales and one ratio.
        self.num_anchors = 2

        # Layers for anchor class, bbox coords and landmark coords.
        self.cls_stride8 = nn.Conv2d(64, 2 * self.num_anchors, 1)
        self.cls_stride16 = nn.Conv2d(64, 2 * self.num_anchors, 1)
        self.cls_stride32 = nn.Conv2d(64, 2 * self.num_anchors, 1)

        self.bbox_stride8 = nn.Conv2d(64, 4 * self.num_anchors, 1)
        self.bbox_stride16 = nn.Conv2d(64, 4 * self.num_anchors, 1)
        self.bbox_stride32 = nn.Conv2d(64, 4 * self.num_anchors, 1)

        self.landmark_stride8 = nn.Conv2d(64, 10 * self.num_anchors, 1)
        self.landmark_stride16 = nn.Conv2d(64, 10 * self.num_anchors, 1)
        self.landmark_stride32 = nn.Conv2d(64, 10 * self.num_anchors, 1)

    def forward(self, x):
        """Forward pass for output predictor.

        Expects `x` to hold one feature map per stride.
        """
        stride8, stride16, stride32 = x

        cls_score8 = self.cls_stride8(stride8)
        cls_score16 = self.cls_stride16(stride16)
        cls_score32 = self.cls_stride32(stride32)

        # Reshape the class scores so we can easily calculate the softmax along
        # the two entries per anchor.
        N, A, H, W = cls_score8.shape
        cls_prob8 = F.softmax(
            cls_score8.view(N, 2, -1, W), dim=1
        ).view(N, A, H, W)
        N, A, H, W = cls_score16.shape
        cls_prob16 = F.softmax(
            cls_score16.view(N, 2, -1, W), dim=1
        ).view(N, A, H, W)
        N, A, H, W = cls_score32.shape
        cls_prob32 = F.softmax(
            cls_score32.view(N, 2, -1, W), dim=1
        ).view(N, A, H, W)

        bbox_pred8 = self.bbox_stride8(stride8)
        bbox_pred16 = self.bbox_stride16(stride16)
        bbox_pred32 = self.bbox_stride32(stride32)

        landmark_pred8 = self.landmark_stride8(stride8)
        landmark_pred16 = self.landmark_stride16(stride16)
        landmark_pred32 = self.landmark_stride32(stride32)

        return [
            cls_prob32,
            bbox_pred32,
            landmark_pred32,

            cls_prob16,
            bbox_pred16,
            landmark_pred16,

            cls_prob8,
            bbox_pred8,
            landmark_pred8,
        ]


class RetinaFace(nn.Module):
    """RetinaFace model with a pseudo-MobileNet backbone.

    Consists of three inner modules: a base network to get a feature pyramid
    from, a pyramid refiner, that adds context and mixes the results a bit, and
    an output predictor that returns the final predictions for class, bounding
    box and landmarks per anchor.

    """

    def __init__(self):
        super().__init__()

        self.base = BaseNetwork()
        self.refiner = PyramidRefiner()
        self.outputs = OutputsPredictor()

    def forward(self, x):
        out = self.base(x)
        out = self.refiner(out)
        out = self.outputs(out)

        return out



import numpy as np
import torch

from cv2 import resize, INTER_LINEAR

from terran import default_device
from terran.checkpoint import get_checkpoint_path
from terran.pose.openpose.model import BodyPoseModel


# TODO: Change names.
map_idx = [
    [31, 32], [39, 40], [33, 34], [35, 36], [41, 42], [43, 44],
    [19, 20], [21, 22], [23, 24], [25, 26], [27, 28], [29, 30],
    [47, 48], [49, 50], [53, 54], [51, 52], [55, 56], [37, 38],
    [45, 46]
]

limbseq = [
    [2, 3], [2, 6], [3, 4], [4, 5], [6, 7], [7, 8], [2, 9],
    [9, 10], [10, 11], [2, 12], [12, 13], [13, 14], [2, 1],
    [1, 15], [15, 17], [1, 16], [16, 18], [3, 17], [6, 18]
]


def load_model():
    model = BodyPoseModel()
    model.load_state_dict(
        torch.load(
            get_checkpoint_path('terran.pose.openpose.OpenPose')
        )
    )
    model.eval()
    return model


def get_keypoints(peaks_by_id, humans, scale=1.0):
    """Build and return the final list of keypoints.

    Also adjust the final keypoint coordinates to the initial image size,
    considering the scaling factor.

    Parameters
    ----------
    peaks_by_id : np.ndarray of shape (N_k, 3)
        Peaks indexed by keypoint id, each row containing location and score.
    humans : np.ndarray of shape (N_d, 20)
        Humans found, with the corresponding keypoint ID at each of the first
        18 locations, plus number of keypoints and scores.
    scale : float
        Scale the image was resized with, needed to adjust final keypoint
        locations.

    Returns
    -------
    List of dicts
        Each dict corresponds to a detected human, and has two entries:
        * `keypoints`: a `np.ndarray`s of size (N_d, 18, 3).
        * `score`: the average score for the human's keypoints.

        `keypoints` is an array containing the keypoint locations for all the
        18 keypoints. The final axis contains the (x, y) coordinates, along
        with a third entry indicating whether the keypoint is present or not.

    """
    detections = []

    if torch.is_tensor(peaks_by_id):
        peaks_by_id = peaks_by_id.cpu().numpy()

    for human in humans:
        keypoints = np.zeros((18, 3), dtype=np.int32)
        for j in range(18):
            peak_id = np.int32(human[j])
            if peak_id != -1:
                y, x = peaks_by_id[peak_id][:2]

                y = (y / scale).astype(np.int32)
                x = (x / scale).astype(np.int32)
                keypoints[j] = (x, y, 1)

        # Average keypoint score: sum of scores over number of keypoints.
        score = human[-2] / human[-1]

        detections.append({
            'keypoints': keypoints,
            'score': score,
        })

    return detections


def resize_images(images, short_side=416):
    H, W = images.shape[1:3]
    scale = short_side / min(H, W)

    new_size = (
        int(W * scale), int(H * scale)
    )

    resized = np.empty(
        (images.shape[0], new_size[1], new_size[0], images.shape[3]),
        dtype=images.dtype
    )
    for idx, image in enumerate(images):
        resize(
            src=image,
            dst=resized[idx],
            dsize=new_size,
            interpolation=INTER_LINEAR,
        )

    return resized, scale


def preprocess_images(images):
    """Preprocess images as required by the pre-trained OpenPose model."""
    # Turn into `BCHW` format.
    out = np.transpose(images, (0, 3, 1, 2))
    out = out.astype(np.float32) / 255.0 - 0.5
    out = torch.as_tensor(out, device=default_device)
    return out


@torch.jit.script
def build_segments(loc_src, loc_dst, num_midpoints: int):
    """Build the `num_midpoint`-point segment between each source and
    destination keypoints of a given limb.

    Parameters
    ----------
    loc_src : torch.Tensor of shape (N_s, 2)
        Locations for the keypoints of one extreme of the limb.
    loc_dst : torch.Tensor of shape (N_d, 2)
        Locations for the keypoints of the other extreme of the limb.
    num_midpoints : int
        Number of points to return per segment.

    Returns
    -------
    torch.Tensor of shape (N_s, N_d, num_midpoints, 2)
        Locations for each of the points between each pair of keypoints.

    """
    count_src = loc_src.shape[0]
    count_dst = loc_dst.shape[0]

    segments = torch.zeros(
        count_src, count_dst, num_midpoints, 2,
        dtype=torch.float32, device=loc_src.device
    )
    for i in range(count_src):
        for j in range(count_dst):
            torch.linspace(
                loc_src[i][0], loc_dst[j][0], steps=num_midpoints,
                out=segments[i, j, :, 0]
            )
            torch.linspace(
                loc_src[i][1], loc_dst[j][1], steps=num_midpoints,
                out=segments[i, j, :, 1]
            )

    return segments


class OpenPose:

    def __init__(self, device=default_device, short_side=184):
        self.device = device
        self.model = load_model().to(self.device)

        self.short_side = short_side

        # Downsampling ratio for the model in use.
        self.downsampling_ratio = 8

        # Keypoint thresholds.
        self.keypoint_threshold = 0.1
        self.thresh_2 = 0.05
        self.human_threshold = 0.4

    def call(self, images):
        """Run the pose estimation model and its postprocessing.

        The preprocessing goes as follow:
        - Get the heatmaps from the model to identify peaks, i.e. candidate
          keypoint locations.
        - Use the PAFs from the model to obtain the limbs. In order to do so,
          for each limb, pick the pair of keypoints that maximizes the line
          integral between both extremes of the limb.
        - Build humans by filling in the needed limbs one by one. In case of a
          conflict (i.e. a limb belongs to two humans), tiebreak by assigning
          to one of them.

        Parameters
        ----------
        images : np.ndarray of shape (N, H, W, C)
            Batch of images to run model through.

        Returns
        -------
        List of humans detected by batch. Each person is specified by a
        `np.ndarry` of shape (18, 3).

        """
        resized, scale = resize_images(images, short_side=self.short_side)
        preprocessed = preprocess_images(resized)

        with torch.no_grad():
            batch_pafs, batch_heatmaps = self.model(preprocessed)

        # TODO: Not adding padding, so it won't detect keypoints in the right
        # border of the image.
        batch_pafs = torch.nn.functional.interpolate(
            batch_pafs,
            scale_factor=self.downsampling_ratio,
            mode='bicubic', align_corners=False,
        )
        batch_heatmaps = torch.nn.functional.interpolate(
            batch_heatmaps,
            scale_factor=self.downsampling_ratio,
            mode='bicubic', align_corners=False,
        )

        batch_objects = []
        for heatmaps, pafs in zip(batch_heatmaps, batch_pafs):

            # Start by obtaining the peak locations for every keypoint, by
            # looking at the returned heatmap.
            num_peaks = 0
            peak_locs = []
            peak_scores = []
            peak_ids = []

            for part in range(18):
                heatmap = heatmaps[part, :, :]

                # Search for local optima. Consider a 1px padding around the
                # map, as we need to make sure it's larger than any surrounding
                # coord.
                peaks_binary = (
                    (heatmap[1:-1, 1:-1] >= heatmap[0:-2, 1:-1])
                    & (heatmap[1:-1, 1:-1] >= heatmap[1:-1, :-2])
                    & (heatmap[1:-1, 1:-1] >= heatmap[2:, 1:-1])
                    & (heatmap[1:-1, 1:-1] >= heatmap[1:-1, 2:])
                    & (heatmap[1:-1, 1:-1] >= self.keypoint_threshold)
                )

                # Add one to the coordinates to account for the 1px padding.
                curr_peaks = torch.nonzero(peaks_binary) + 1
                curr_num_peaks = curr_peaks.shape[0]

                curr_peak_ids = torch.arange(
                    num_peaks, num_peaks + curr_num_peaks
                )
                curr_scores = heatmap[curr_peaks[:, 0], curr_peaks[:, 1]]

                peak_locs.append(curr_peaks)
                peak_scores.append(curr_scores)
                peak_ids.append(curr_peak_ids)

                num_peaks += curr_num_peaks

            # Now build the limbs out of the detected keypoints.

            # `all_connections` will be a list containing number-of-limbs
            # elements, each with a list of connections for said limb. Within
            # those list, the connection is represented by a 3-tuple of the
            # peak IDs of both extremes and the limb score.
            all_connections = []
            missing_limbs = []
            num_midpoints = 10

            for limb_id in range(len(map_idx)):
                # Calculate line integrals between each source and destination
                # keypoints for the current limb. Get the segments between the
                # points, the directions, and multiply by the values of the
                # vector field.
                paf = pafs[[x - 19 for x in map_idx[limb_id]], :, :]

                # Keypoint IDs for limb.
                kpid_src = limbseq[limb_id][0] - 1
                kpid_dst = limbseq[limb_id][1] - 1

                loc_src = peak_locs[kpid_src]
                loc_dst = peak_locs[kpid_dst]

                count_src = loc_src.shape[0]
                count_dst = loc_dst.shape[0]

                if count_src == 0 or count_dst == 0:
                    missing_limbs.append(limb_id)
                    all_connections.append(())
                    continue

                directions = (
                    loc_dst.reshape(1, -1, 2) - loc_src.reshape(-1, 1, 2)
                ).type(torch.float32)
                norms = torch.norm(directions, dim=2)
                directions = directions / norms[..., None]

                # Cast it to `torch.long` so we transform it into PAF
                # locations, for easy evaluation.
                segments = build_segments(
                    loc_src, loc_dst, num_midpoints
                ).type(torch.long)

                midpoint_scores = torch.mul(
                    paf[
                        :, segments[..., 0], segments[..., 1]
                    ].permute(3, 0, 1, 2),
                    # Flip the directions, as the network output is in `(x, y)`
                    # and our `directions` vector in `(y, x)`.
                    torch.flip(directions.permute(2, 0, 1), dims=(0,))
                ).sum(dim=1)

                # Score with length regularization.
                # TODO: Where does this heuristic come from? Does it have to be
                # like this? Why heigth and not width?
                reg_scores = (
                    midpoint_scores.sum(dim=0) / midpoint_scores.shape[0]
                    + torch.clamp(0.5 * pafs.shape[1] / norms - 1, max=0)
                )

                criterion_1 = (
                    (
                        midpoint_scores > self.thresh_2
                    ).sum(dim=0) > 0.8 * num_midpoints
                )
                criterion_2 = (reg_scores > 0)

                matching = (criterion_1 & criterion_2).nonzero()
                matching_scores = reg_scores[matching[:, 0], matching[:, 1]]

                connections = []
                seen = set()

                # Perform the actual matching by greedily building connections
                # with the highest-scoring pairs.
                # TODO: Improve from this part onwards..
                for match in matching.cpu().numpy()[
                    np.argsort(-matching_scores.cpu().numpy())
                ]:
                    i, j = match
                    s = reg_scores[i, j].cpu().numpy()
                    if i not in seen and j not in seen:
                        connections.append(
                            np.array([
                                peak_ids[kpid_src][i],
                                peak_ids[kpid_dst][j],
                                s
                            ])
                        )

                        if len(connections) >= min(count_src, count_dst):
                            break

                        seen.add(i)
                        seen.add(j)

                if connections:
                    connections = np.stack(connections)
                else:
                    connections = np.zeros((0, 3))

                all_connections.append(connections)

            # Build each person by progressively merging the found limbs.

            # Peak location and score, indexed by ID.
            peaks_by_id = np.array([
                tuple(p) + (sc,)
                for pks, scs in zip(peak_locs, peak_scores)
                for p, sc in zip(pks, scs)
            ])

            # First 18 entries are the keypoints, following one is the number
            # of keypoints, and the last one is the sum of the connection
            # scores.
            humans = np.ones((0, 20)) * -1

            for limb_id in range(len(map_idx)):
                if limb_id in missing_limbs:
                    continue

                peak_src = all_connections[limb_id][:, 0]
                peak_dst = all_connections[limb_id][:, 1]
                kpid_src, kpid_dst = np.array(limbseq[limb_id]) - 1

                for conn_idx in range(len(all_connections[limb_id])):

                    # Check for matches of the current connection with existing
                    # humans. By construction, up to two matches may occur.
                    matched_with = []
                    for human_idx, human in enumerate(humans):
                        if (
                            human[kpid_src] == peak_src[conn_idx]
                            or human[kpid_dst] == peak_dst[conn_idx]
                        ):
                            matched_with.append(human_idx)

                    if len(matched_with) == 1:
                        # There was a match with only one human. Add the
                        # destination keypoint to it. We don't add the source
                        # keypoint, as it must have been added already, due to
                        # the limbs being iterated in order.
                        human = humans[matched_with[0]]
                        if human[kpid_dst] != peak_dst[conn_idx]:
                            human[kpid_dst] = peak_dst[conn_idx]
                            human[-1] += 1
                            human[-2] += (
                                peaks_by_id[peak_dst[conn_idx].astype(int), 2]
                                + all_connections[limb_id][conn_idx][2]
                            )

                    elif len(matched_with) == 2:
                        human_1_idx, human_2_idx = matched_with
                        human_1 = humans[human_1_idx]
                        human_2 = humans[human_2_idx]

                        # Check the number of keypoints present in both humans
                        # matched. If they're non-overlapping, then we're
                        # connecting two human parts, so merge them into a
                        # single human.
                        membership = (
                            (human_1 >= 0).astype(int)
                            + (human_2 >= 0).astype(int)
                        )[:-2]
                        non_overlapping = (
                            len(np.flatnonzero(membership == 2)) == 0
                        )
                        if non_overlapping:
                            # Merge both humans and sum the number of keypoints
                            # and scores.
                            # The `+1` because absence of the limb is marked as
                            # `-1`.
                            human_1[:-2] += (human_2[:-2] + 1)
                            human_1[-2:] += human_2[-2:]
                            human_1[-2] += (
                                all_connections[limb_id][conn_idx][2]
                            )
                            humans = np.delete(humans, human_2_idx, 0)
                        else:
                            # There's a conflict, as the people overlap.
                            # Tiebreak by adding the connection into the first
                            # one.
                            human_1[kpid_dst] = peak_dst[conn_idx]
                            human_1[-1] += 1
                            human_1[-2] += (
                                peaks_by_id[peak_dst[conn_idx].astype(int), 2]
                                + all_connections[limb_id][conn_idx][2]
                            )

                    elif not matched_with and limb_id < 17:
                        # New human found, add a row.
                        human = np.ones(20) * -1
                        human[kpid_src] = peak_src[conn_idx]
                        human[kpid_dst] = peak_dst[conn_idx]
                        human[-1] = 2
                        human[-2] = sum(
                            peaks_by_id[
                                all_connections[limb_id][
                                    conn_idx, :2
                                ].astype(int),
                                2
                            ]
                        ) + all_connections[limb_id][conn_idx][2]
                        humans = np.vstack([humans, human])

            # Delete all detected humans with less than four keypoints and an
            # average keypoint score less than `0.4`.
            to_delete = []
            for human_idx, human in enumerate(humans):
                num_keypoints = human[-1]
                avg_score = human[-2] / human[-1]
                if num_keypoints < 4 or avg_score < self.human_threshold:
                    to_delete.append(human_idx)
            humans = np.delete(humans, to_delete, axis=0)

            # Build the final list of keypoints.
            batch_objects.append(
                get_keypoints(peaks_by_id, humans, scale=scale)
            )

        return batch_objects



import numpy as np

from filterpy.kalman import KalmanFilter
from scipy.optimize import linear_sum_assignment

from terran.face.detection import Detection, face_detection


def linear_assignment(cost_matrix):
    """Implement the linear assignment as in Scikit Learn v0.21"""
    return np.transpose(np.asarray(linear_sum_assignment(cost_matrix)))


def iou(bbox_1, bbox_2):
    """Computes intersection over union between two bounding boxes.

    Parameters
    ----------
    bbox_1 : np.ndarray
        First bounding box, of the form (x_min, y_min, x_max, y_max).
    bbox_2 : np.ndarray
        Second bounding box, of the form (x_min, y_min, x_max, y_max).

    Returns
    -------
    float
        Intersection over union value between both bounding boxes.

    """
    x_min = np.maximum(bbox_1[0], bbox_2[0])
    y_min = np.maximum(bbox_1[1], bbox_2[1])
    x_max = np.minimum(bbox_1[2], bbox_2[2])
    y_max = np.minimum(bbox_1[3], bbox_2[3])
    width = np.maximum(0.0, x_max - x_min)
    height = np.maximum(0.0, y_max - y_min)
    intersection = width * height

    return (
        intersection
    ) / (
        (bbox_1[2] - bbox_1[0]) * (bbox_1[3] - bbox_1[1])
        + (bbox_2[2] - bbox_2[0]) * (bbox_2[3] - bbox_2[1])
        - intersection
    )


def corners_to_center(bbox):
    """Changes bounding box from corner-based specification to center-based.

    Parameters
    ---------
    bbox : np.ndarray
        Bounding box of the form (x_min, y_min, x_max, y_max).

    Returns
    -------
    np.ndarray
        Same bounding box, but of the form (x, y, area, ratio).

    """
    width = bbox[2] - bbox[0]
    height = bbox[3] - bbox[1]

    x = bbox[0] + width / 2.0
    y = bbox[1] + height / 2.0

    area = width * height
    ratio = width / height

    return np.array([
        x, y, area, ratio
    ]).reshape((4, 1))


def center_to_corners(bbox):
    """Changes bounding box from corner-based specification to center-based.

    Parameters
    ---------
    bbox : np.ndarray
        Bounding box of the form (x, y, area, ratio).

    Returns
    -------
    np.ndarray
        Same bounding box, but of the form (x_min, y_min, x_max, y_max).

    """
    width = np.sqrt(bbox[2] * bbox[3])
    height = bbox[2] / width

    return np.concatenate([
        bbox[0] - width / 2.0,
        bbox[1] - height / 2.0,
        bbox[0] + width / 2.0,
        bbox[1] + height / 2.0
    ])


class KalmanTracker:
    """Tracker for individual face.

    Maintains an internal state by way of a Kalman filter applied to the face's
    bounding boxes.

    The Kalman filter used tracks the bounding box's center point, scale and
    ratio, and assumes a constant velocity of the box, and a constant ratio for
    the faces: we don't try to estimate the velocity of the ratio, as it's
    bound to be incorrect in a very short time.

    """

    # Count of all trackers instantiated, to keep ascending names for tracks.
    count = 0

    def __init__(self, face):
        """Initializes a tracker using an initial detected face.

        Parameters
        ----------
        face : dict
            Face to initialize tracker to, as returned by a `Detection`
            instance.

        """
        # 4 measurements plus 3 velocities. We don't keep a velocity for the
        # ratio of the bounding box.
        self.kf = KalmanFilter(dim_x=7, dim_z=4)

        self.kf.F = np.array([
            [1, 0, 0, 0, 1, 0, 0],
            [0, 1, 0, 0, 0, 1, 0],
            [0, 0, 1, 0, 0, 0, 1],
            [0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 0, 1],
        ])

        self.kf.H = np.array([
            [1, 0, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0],
        ])

        self.kf.R[2:, 2:] *= 10.0

        # Give high uncertainty to the unobservable initial velocities.
        self.kf.P[4:, 4:] *= 1000.0
        self.kf.P *= 10.0

        self.kf.Q[-1, -1] *= 0.01
        self.kf.Q[4:, 4:] *= 0.01

        self.kf.x[:4] = corners_to_center(face['bbox'])

        self.hits = 0
        self.time_since_update = 0

        self.id = KalmanTracker.count
        KalmanTracker.count += 1

    def update(self, face):
        """Updates the state vector with observed face.

        Parameters
        ----------
        face : dict
            Face to update the tracker with, an observation that's already been
            assigned to the predicted trajectory of the tracker.

        """
        self.time_since_update = 0
        self.hits += 1
        self.kf.update(corners_to_center(face['bbox']))

    def predict(self):
        """Advances the state vector and returns the predicted bounding box
        estimate.

        Returns
        -------
        np.ndarray
            Predicted trajectory of the tracker, in (x_min, y_min, x_max,
            y_max) format.

        """
        # If the size of the bounding box is negative, nullify the velocity.
        if (self.kf.x[6] + self.kf.x[2]) <= 0:
            self.kf.x[6] *= 0.0

        self.kf.predict()
        self.time_since_update += 1

        return center_to_corners(self.kf.x)


def associate_detections_to_trackers(faces, trackers, iou_threshold=0.3):
    """Assigns detections to tracked faces.

    Parameters
    ----------
    faces : list
        Observed faces, as returned by a Detection class.
    trackers : np.ndarray of size (T,)
        Positions for the `T` existing trackers.
    iou_threshold : float
        Threshold of IoU value for considering two boxes a match.

    Returns
    -------
    list of np.ndarray
        Matched, unmatched faces and unmatched trackers.

    """

    if not len(trackers):
        return (
            np.empty((0, 2), dtype=int),
            np.arange(len(faces)),
            np.empty((0, 5), dtype=int),
        )

    iou_matrix = np.zeros(
        (len(faces), len(trackers)),
        dtype=np.float32
    )
    for face_idx, face in enumerate(faces):
        for track_idx, track in enumerate(trackers):
            iou_matrix[face_idx, track_idx] = iou(face['bbox'], track)

    # Use the hungarian method to get least-cost assignment. Returns an array
    # of size (min(len(faces), len(trackers)), 2), with the (row, col) indices
    # of the matches.
    matched_indices = linear_assignment(-iou_matrix)

    unmatched_faces = []
    for face_idx, face in enumerate(faces):
        if face_idx not in matched_indices[:, 0]:
            unmatched_faces.append(face_idx)

    unmatched_trackers = []
    for track_idx, track in enumerate(trackers):
        if track_idx not in matched_indices[:, 1]:
            unmatched_trackers.append(track_idx)

    # Filter out matched with low IOU.
    matches = []
    for face_idx, track_idx in matched_indices:
        if iou_matrix[face_idx, track_idx] < iou_threshold:
            unmatched_faces.append(face_idx)
            unmatched_trackers.append(track_idx)
        else:
            matches.append(
                np.array([face_idx, track_idx], dtype=int)
            )

    if not matches:
        matches = np.empty((0, 2), dtype=int)
    else:
        matches = np.stack(matches)

    return (
        matches, np.array(unmatched_faces), np.array(unmatched_trackers)
    )


class Sort:
    """SORT tracker for very simple, appearence-agnostic tracking.

    Implements a multiple object tracker based on `Simple Online and Realtime
    Tracking`_ focused on building *face tracks* out of detections (thus
    performing tracking-by-detection).

    The tracking performed by this class has only one objective: attach an
    identity to every detection passed to it, or `None` if no identity was
    found or yet constructed for it.

    This means the class will not:
    * Smooth out the bounding boxes: observations are returned as-is.
    * Interpolate a face whenever there's a missing observation in between.

    If you need to do get the smoothed out observations, your best bet is to
    modify the `self.update()` method to return the tracker predictions instead
    of the observations. We might make it a configurable at a later point.

    .. _Simple Online and Realtime Tracking: https://arxiv.org/abs/1602.00763

    """

    def __init__(self, max_age=1, min_hits=3, return_unmatched=False):
        """Initialize SORT instance with its configuration.

        The recommended way of setting the time-based parameters is with
        respect to the framerate of the video.

        Parameters
        ----------
        max_age : int
            Maximum number of steps after which an unmatched tracker is going
            to be destroyed.
        min_hits : int
            Minimum number of steps until a tracker is considered confirmed
            (and thus, its identity returned in faces).
        return_unmatched : bool
            Whether to return faces with no track attached (default: False).

        """
        self.max_age = max_age
        self.min_hits = min_hits
        self.return_unmatched = return_unmatched

        self.trackers = []
        self.frame_count = 0

    def update(self, faces):
        """Update the tracker with new faces.

        This function should be called on every frame, even if no faces are
        detected.

        Parameters
        ----------
        faces : list
            List of faces dicts, as returned by a `Detection` instance.

        Returns
        -------
        list of dicts
            Same list of face dicts, but with an extra `track` field specifying
            the identity of the face. This field will either be `None` if the
            face wasn't matched to any underlying track, or an `int` if it was.

            If `self.return_unmatched` is `False`, all faces will have a
            `track`, or else they'll get filtered.

        """
        self.frame_count += 1

        # Get predicted locations from existing trackers.
        to_delete = []
        tracks = np.zeros((len(self.trackers), 4))
        for track_idx, track in enumerate(tracks):

            position = self.trackers[track_idx].predict()
            track[:] = position

            # A tracker might have gone to infinity in its extrapolation.
            if np.any(np.isnan(position)):
                to_delete.append(track_idx)

        tracks = np.ma.compress_rows(
            np.ma.masked_invalid(tracks)
        )

        for t in reversed(to_delete):
            self.trackers.pop(t)

        (
            matched, unmatched_faces, unmatched_tracks
        ) = associate_detections_to_trackers(faces, tracks)

        # Faces to return, augmented with the track ID, whenever available.
        augmented_faces = []

        # Update matched trackers with assigned detections.
        for track_idx, track in enumerate(self.trackers):
            if track_idx not in unmatched_tracks:
                face_idx = int(
                    matched[np.where(matched[:, 1] == track_idx)[0], 0]
                )
                track.update(faces[face_idx])

                # Add the track ID to the `track` field only if the track has
                # been confirmed.
                track_id = track.id if (
                    track.hits >= self.min_hits
                    or self.frame_count <= self.min_hits
                ) else None

                augmented_faces.append(
                    {'track': track_id, **faces[face_idx]}
                )

        # Create and initialize new trackers for unmatched detections.
        for face_idx in unmatched_faces:
            track = KalmanTracker(faces[face_idx])
            self.trackers.append(track)

            # Just created: only case we should return it right away is if
            # there are no minimum amount of hits required.
            track_id = track.id if self.min_hits == 0 else None
            augmented_faces.append(
                {'track': track_id, **faces[face_idx]}
            )

        # Filter out the faces without a confirmed tracker attached.
        if not self.return_unmatched:
            augmented_faces = [
                face for face in augmented_faces
                if face['track'] is not None
            ]

        # Finally, clean up dead tracks.
        self.trackers = [
            track for track in self.trackers
            if track.time_since_update <= self.max_age
        ]

        return augmented_faces


class FaceTracking:
    """Object for performing face tracking.

    This object is meant to be used as a substitute to a `Detection` object,
    behaving exactly the same way except for having an extra `track` field in
    the face dictionaries.

    The object will only encapsulate and call the detector and tracker objects
    used, offer a :meth:`__call__`-based interface. That is, it's simply a
    container for the main :class:`Sort` class.

    """

    def __init__(self, detector=None, tracker=None):
        self.detector = detector
        self.tracker = tracker

    def __call__(self, frames):
        """Performs face tracking on `images`.

        The face detection itself will be done by the `self.detector` object,
        while the tracking by the `self.tracker` object.

        Parameters
        ----------
        frames : list of numpy.ndarray or numpy.ndarray
            Frames to perform face tracking on.

        Returns
        -------
        list of list of dicts, or list dict
            List of dictionaries containing face data for a single image, or a
            list of these entries thereof.

            Each entry is of the form::

                {
                    'bbox': [x_min, y_min, x_max, y_max],
                    'landmarks': ...,  # Array of shape (5, 2).
                    'track': ...,  # Either an `int` or `None`.
                    'score': ... # Confidence score.
                }

        """
        expanded = False
        if not isinstance(frames, list) and len(frames.shape) == 3:
            expanded = True
            frames = frames[0]

        faces_per_frame = []
        detections_per_frame = self.detector(frames)
        for frame, detections in zip(frames, detections_per_frame):
            faces_per_frame.append(
                self.tracker.update(detections)
            )

        return faces_per_frame[0] if expanded else faces_per_frame


def face_tracking(
    *, video=None, max_age=None, min_hits=None, detector=None,
    return_unmatched=False,
):
    """Default entry point to face tracking.

    This is a factory for an underlying :class:`FaceTracking` instance,
    which will be tasked with keeping the state of the different identities
    available.

    Once you create it, you can treat the resulting object as if it was an
    instance of the :class:`terran.face.detection.Detection` class, but focused
    on working in same-size batches of frames, and returning an additional
    field on the faces corresponding to the identity, or track.

    The tracking utilities provided focus on filtering observations *only*.
    No smoothing nor interpolation will be performed, so the results you
    obtained can be traced back to detections of the detector passed on to it.
    This is meant as a building block from which to do more detailed face
    recognition over videos.

    Parameters
    ----------
    video : terran.io.reader.Video
        Video to derive `max_age` and `min_hits` from. The first value will be
        one video of the second, while the latter will be 1/5th of a second.

        When those values are specified as well, they'll have precedence.
    max_age : int
        Maximum number of frames to keep identities around for after no
        appearance.
    min_hits : int
        Minimum number of observations required for an identity to be returned.

        For instance, if `min_hits` is 6, it means that only after a face is
        detected six times will it be returned on prediction. This is, in
        essence, adding *latency* to the predictions. So consider decreasing
        this value if you care more about latency than any possible noise you
        may get due to short-lived faces.

        You can also get around this latency by specifying `return_unmatched`
        value of `True`, but in that case, returned faces will *not* have an
        identity associated.
    detector : terran.face.detection.Detection
        Face detector to get observations from. Default is using Terran's
        default face detection class.
    return_unmatched : boolean
        Whether to return observations (faces) that don't have a matched
        identity or not.

    """
    # Default values for SORT assume a 30 fps video.
    max_age_ = 30
    min_hits_ = 6

    # If we receive a video or any of the parameters are specified, substitute
    # the defaults.
    if video is not None:
        max_age_ = video.framerate
        min_hits_ = video.framerate // 5

    if max_age is None:
        max_age = max_age_
    if min_hits is None:
        min_hits = min_hits_

    # Validate that we received a valid detector, or fall back to the default
    # one if none was specified.
    if detector is None:
        detector = face_detection
    elif not isinstance(detector, Detection):
        raise ValueError(
            '`detector` must be an instance of `terran.face.Detection`.'
        )

    sort = Sort(
        max_age=video.framerate,
        min_hits=video.framerate // 5,
        return_unmatched=return_unmatched,
    )

    return FaceTracking(detector=detector, tracker=sort)



import numpy as np
import sys

from PIL import Image, ImageDraw, ImageFont

from terran.pose import Keypoint
from terran.vis import (
    FACE_COLORMAP, POSE_CONNECTIONS, POSE_CONNECTION_COLORS,
    POSE_KEYPOINT_COLORS,
)


def get_font():
    """Attempts to retrieve a reasonably-looking TTF font from the system.

    We don't make much of an effort, but it's what we can reasonably do without
    incorporating additional dependencies for this task.
    """
    if sys.platform == 'win32':
        font_names = ['Arial']
    elif sys.platform in ['linux', 'linux2']:
        font_names = ['DejaVuSans-Bold', 'DroidSans-Bold']
    elif sys.platform == 'darwin':
        font_names = ['Menlo', 'Helvetica']

    font = None
    for font_name in font_names:
        try:
            font = ImageFont.truetype(font_name)
            break
        except IOError:
            continue

    return font


SYSTEM_FONT = get_font()


def draw_label(draw, coords, text, color, scale=1):
    """Draw a box with the label and probability."""
    # Attempt to get a native TTF font. If not, use the default bitmap font.
    global SYSTEM_FONT
    if SYSTEM_FONT:
        label_font = SYSTEM_FONT.font_variant(size=round(16 * scale))
    else:
        label_font = ImageFont.load_default()

    text = str(text)  # `text` may not be a string.

    # We want the probability font to be smaller, so we'll write the label in
    # two steps.
    text_w, text_h = label_font.getsize(text)

    # Get margins to manually adjust the spacing. The margin goes between each
    # segment (i.e. margin, label, margin, prob, margin).
    margin_w, margin_h = label_font.getsize('M')
    margin_w *= 0.2
    _, full_line_height = label_font.getsize('Mq')

    # Draw the background first, considering all margins and the full line
    # height.
    background_coords = [
        coords[0],
        coords[1],
        coords[0] + text_w + + 3 * margin_w,
        coords[1] + full_line_height * 1.15,
    ]
    draw.rectangle(background_coords, fill=color + (255,))

    # Then write the two pieces of text.
    draw.text([
        coords[0] + margin_w,
        coords[1],
    ], text, font=label_font)


def draw_marker(draw, coords, color=(255, 0, 0), scale=1):
    """Draw a marker on `draw` at `coords`.

    The marker itself is a rectangle with rounded corners.
    """
    scale = int(3 * scale)
    outline = color + (255,)
    draw.rectangle(list(coords), outline=outline, width=scale)


def vis_faces(image, faces, scale=1.0):
    """Draw boxes over the detected faces for the given image.

    Parameters
    ----------
    image : np.ndarray representing an image
        Image to draw faces over.
    faces : dict or list of dicts (from `face_detection` or `face_tracking`)
        Faces to draw on `image`. The expected format is the one returned from
        `face_detection` or `face_tracking`, with two optional extra fields:

        - ``text`` (str): Text to be written next to the box.
        - ``name`` (str): Name associated to the face, in order to make the
          color used for the box fixed.

        If available, the ``track`` field will be used as the default value
        for the two values above, if they aren't specified.

    Returns
    -------
    np.ndarray
        Copy of `image` with the faces drawn over.

    """
    if not (isinstance(faces, list) or isinstance(faces, tuple)):
        faces = [faces]

    image = Image.fromarray(image)
    draw = ImageDraw.Draw(image, 'RGBA')

    for face in faces:
        # Get the name and text for the current face.
        face_name = face.get('name') or face.get('track')
        if face.get('text') is not None:
            face_text = face['text']
        elif face.get('track') is not None:
            face_text = f"#{face['track']}"
        else:
            face_text = None

        color = tuple(FACE_COLORMAP(face_name))
        draw_marker(draw, face['bbox'], color=color, scale=scale)

        if face_text is not None:
            draw_label(
                draw, face['bbox'][:2], face_text, color, scale=scale
            )

    return np.asarray(image)


def draw_keypoints(draw, keypoints, scale=1.0):
    scale = int(scale * 4)

    for keypoint in keypoints:
        for idx, (x, y, is_present) in enumerate(keypoint['keypoints']):
            if not is_present:
                continue

            color = tuple(POSE_KEYPOINT_COLORS[Keypoint(idx)]) + (225,)
            draw.ellipse([
                x - int(3 * scale / 2), y - int(3 * scale / 2),
                x + int(3 * scale / 2), y + int(3 * scale / 2),
            ], fill=color)


def draw_limbs(draw, keypoints, scale=1.0):
    scale = int(scale * 8)

    for keypoint in keypoints:
        kps = keypoint['keypoints']
        for idx, (conn_src, conn_dst) in enumerate(POSE_CONNECTIONS):
            x_src, y_src, src_present = kps[conn_src.value]
            x_dst, y_dst, dst_present = kps[conn_dst.value]

            # Ignore limbs for which one of the keypoints is missing.
            if not (src_present and dst_present):
                continue

            color = tuple(POSE_CONNECTION_COLORS[idx]) + (180,)
            draw.line([x_src, y_src, x_dst, y_dst], fill=color, width=scale)


def vis_poses(image, poses, scale=1.0):
    """Draw boxes over the detected poses for the given image.

    Parameters
    ----------
    image : np.ndarray representing an image.
        Image to draw faces over.
    poses : dict or list of dicts, as returned by `pose_estimation`
        Poses to draw on `image`. The expected format is the one returned from
        `pose_estimation`.

    Returns
    -------
    np.ndarray
        Copy of `image` with the poses drawn over.

    """
    if not (isinstance(poses, list) or isinstance(poses, tuple)):
        poses = [poses]

    image = Image.fromarray(image)
    draw = ImageDraw.Draw(image, 'RGBA')

    draw_limbs(draw, poses, scale=scale)
    draw_keypoints(draw, poses, scale=scale)

    return np.asarray(image)
